
from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class BLOCKAType(TREElement):
    def __init__(self, value):
        super(BLOCKAType, self).__init__()
        self.add_field('BLOCK_INSTANCE', 's', 2, value)
        self.add_field('N_GRAY', 's', 5, value)
        self.add_field('L_LINES', 's', 5, value)
        self.add_field('LAYOVER_ANGLE', 's', 3, value)
        self.add_field('SHADOW_ANGLE', 's', 3, value)
        self.add_field('RESERVED_001', 's', 16, value)
        self.add_field('FRLC_LOC', 's', 21, value)
        self.add_field('LRLC_LOC', 's', 21, value)
        self.add_field('LRFC_LOC', 's', 21, value)
        self.add_field('FRFC_LOC', 's', 21, value)
        self.add_field('RESERVED_002', 's', 5, value)


class BLOCKA(TREExtension):
    _tag_value = 'BLOCKA'
    _data_type = BLOCKAType



"""
Common use elements for the NGA modified RDE/AFRL labeling definition
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"


import numpy
from datetime import date, datetime
from typing import Optional

from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import DateTimeDescriptor, FloatDescriptor, \
    SerializableDescriptor, StringEnumDescriptor, StringDescriptor

from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.io.complex.sicd_elements.blocks import XYZType
from .base import DEFAULT_STRICT


class DateRangeType(Serializable, Arrayable):
    """
    A range of dates with resolution of 1 day
    """
    _fields = ('Begin', 'End')
    _required = _fields
    # descriptors
    Begin = DateTimeDescriptor(
        'Begin', _required, strict=DEFAULT_STRICT, numpy_datetime_units='D',
        docstring="Begin date of the data collection.")  # type: Optional[numpy.datetime64]
    End = DateTimeDescriptor(
        'End', _required, strict=DEFAULT_STRICT, numpy_datetime_units='D',
        docstring="End date of the data collection.")  # type: Optional[numpy.datetime64]

    def __init__(self, Begin=None, End=None, **kwargs):
        """
        Parameters
        ----------
        Begin : None|numpy.datetime64|str|datetime|date
        End : None|numpy.datetime64|str|datetime|date
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Begin = Begin
        self.End = End
        super(DateRangeType, self).__init__(**kwargs)

    def get_array(self, dtype='datetime64[D]'):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype
            data type of the return

        Returns
        -------
        numpy.ndarray
            data array
        """

        return numpy.array([self.Begin, self.End], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Begin, End]

        Returns
        -------
        DateRangeType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(Begin=array[0], End=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class DateTimeRangeType(Serializable, Arrayable):
    """
    A range of dates with resolution of 1 day
    """
    _fields = ('Begin', 'End')
    _required = _fields
    # descriptors
    Begin = DateTimeDescriptor(
        'Begin', _required, strict=DEFAULT_STRICT, numpy_datetime_units='s',
        docstring="Begin date/time of the data collection.")  # type: Optional[numpy.datetime64]
    End = DateTimeDescriptor(
        'End', _required, strict=DEFAULT_STRICT, numpy_datetime_units='s',
        docstring="End date/time of the data collection.")  # type: Optional[numpy.datetime64]

    def __init__(self, Begin=None, End=None, **kwargs):
        """
        Parameters
        ----------
        Begin : None|numpy.datetime64|str|datetime|date
        End : None|numpy.datetime64|str|datetime|date
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Begin = Begin
        self.End = End
        super(DateTimeRangeType, self).__init__(**kwargs)

    def get_array(self, dtype='datetime64[D]'):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype
            data type of the return

        Returns
        -------
        numpy.ndarray
            data array
        """

        return numpy.array([self.Begin, self.End], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Begin, End]

        Returns
        -------
        DateTimeRangeType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(Begin=array[0], End=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class RangeCrossRangeType(Serializable, Arrayable):
    """
    A range and cross range attribute container
    """
    _fields = ('Range', 'CrossRange')
    _required = _fields
    _numeric_format = {key: '0.17G' for key in _fields}
    # descriptors
    Range = FloatDescriptor(
        'Range', _required, strict=True, docstring='The Range attribute.')  # type: float
    CrossRange = FloatDescriptor(
        'CrossRange', _required, strict=True, docstring='The Cross Range attribute.')  # type: float

    def __init__(self, Range=None, CrossRange=None, **kwargs):
        """
        Parameters
        ----------
        Range : float
        CrossRange : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Range, self.CrossRange = Range, CrossRange
        super(RangeCrossRangeType, self).__init__(**kwargs)

    def get_array(self, dtype='float64'):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Range, CrossRange]
        """

        return numpy.array([self.Range, self.CrossRange], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Range, CrossRange]

        Returns
        -------
        RangeCrossRangeType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(Range=array[0], CrossRange=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class RowColDoubleType(Serializable, Arrayable):
    _fields = ('Row', 'Col')
    _required = _fields
    _numeric_format = {key: '0.17G' for key in _fields}
    # Descriptors
    Row = FloatDescriptor(
        'Row', _required, strict=True, docstring='The Row attribute.')  # type: float
    Col = FloatDescriptor(
        'Col', _required, strict=True, docstring='The Column attribute.')  # type: float

    def __init__(self, Row=None, Col=None, **kwargs):
        """
        Parameters
        ----------
        Row : float
        Col : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Row, self.Col = Row, Col
        super(RowColDoubleType, self).__init__(**kwargs)

    def get_array(self, dtype='float64'):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Row, Col]
        """

        return numpy.array([self.Row, self.Col], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Row, Col]

        Returns
        -------
        RowColDoubleType
        """
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(Row=array[0], Col=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class LatLonEleType(Serializable, Arrayable):
    """A three-dimensional geographic point in WGS-84 coordinates."""
    _fields = ('Lat', 'Lon', 'Ele')
    _required = _fields
    _numeric_format = {'Lat': '0.17G', 'Lon': '0.17G', 'Ele': '0.17G'}
    # descriptors
    Lat = FloatDescriptor(
        'Lat', _required, strict=True,
        docstring='The latitude attribute. Assumed to be WGS-84 coordinates.'
    )  # type: float
    Lon = FloatDescriptor(
        'Lon', _required, strict=True,
        docstring='The longitude attribute. Assumed to be WGS-84 coordinates.'
    )  # type: float
    Ele = FloatDescriptor(
        'Ele', _required, strict=True,
        docstring='The Height Above Ellipsoid (in meters) attribute. '
                  'Assumed to be WGS-84 coordinates.')  # type: float

    def __init__(self, Lat=None, Lon=None, Ele=None, **kwargs):
        """
        Parameters
        ----------
        Lat : float
        Lon : float
        Ele : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Lat = Lat
        self.Lon = Lon
        self.Ele = Ele
        super(LatLonEleType, self).__init__(Lat=Lat, Lon=Lon, **kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the data.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            data type of the return

        Returns
        -------
        numpy.ndarray
            data array with appropriate entry order
        """

        return numpy.array([self.Lat, self.Lon, self.Ele], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Lat, Lon, Ele]

        Returns
        -------
        LatLonEleType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 3:
                raise ValueError('Expected array to be of length 3, and received {}'.format(array))
            return cls(Lat=array[0], Lon=array[1], Ele=array[2])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class ProjectionPerturbationType(Serializable):
    """
    Basic information required for SICD/SIDD projection model perturbation.
    """
    
    _fields = ('CoordinateFrame', 'DeltaArp', 'DeltaVarp', 'DeltaRange')
    _required = ('CoordinateFrame', )
    _numeric_format = {'Lat': '0.17G', }
    CoordinateFrame = StringEnumDescriptor(
        'CoordinateFrame', {'ECF', 'RIC_ECI', 'RIC_ECF'}, _required)  # type: str
    DeltaArp = SerializableDescriptor(
        'DeltaArp', XYZType, _required)  # type: XYZType
    DeltaVarp = SerializableDescriptor(
        'DeltaVarp', XYZType, _required)  # type: XYZType
    DeltaRange = FloatDescriptor(
        'DeltaRange', _required)  # type: float

    def __init__(self, CoordinateFrame=None, DeltaArp=None, DeltaVarp=None, DeltaRange=None, **kwargs):
        """
        Parameters
        ----------
        CoordinateFrame : str
        DeltaArp : None|XYZType|numpy.ndarray|list|tuple
        DeltaVarp : None|XYZType|numpy.ndarray|list|tuple
        DeltaRange : None|float
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        self.CoordinateFrame = CoordinateFrame
        self.DeltaArp = DeltaArp
        self.DeltaVarp = DeltaVarp
        self.DeltaRange = DeltaRange

        super(ProjectionPerturbationType, self).__init__(**kwargs)

    def set_coa_projection(self, structure):
        """
        Sets the sicd or sidd coa_projection property, as appropriate.

        Parameters
        ----------
        structure : SICDType|SIDDType1|SIDDType2
        """

        if not isinstance(structure, (SICDType, SIDDType1, SIDDType2)):
            raise TypeError('Requires input of type SICDType or SIDDType, got {}'.format(type(structure)))

        structure.define_coa_projection(
            delta_arp=None if self.DeltaArp is None else self.DeltaArp.get_array(dtype='float64'),
            delta_varp=None if self.DeltaVarp is None else self.DeltaVarp.get_array(dtype='float64'),
            range_bias=self.DeltaRange,
            adj_params_frame=self.CoordinateFrame,
            override=True)


class LabelSourceType(Serializable):
    _fields = ('SourceType', 'SourceID', 'Description')
    _required = ('SourceType', )
    SourceType = StringEnumDescriptor(
        'SourceType', {
            'Ground Truth', 'Analyst Truth', 'Algorithm Truth', 'Other', 'Unknown'},
        _required,
        docstring='The source type of the labeling effort')  # type: str
    SourceID = StringDescriptor(
        'SourceID', _required,
        docstring='The "ID" of the labeling source. '
                  'This should be populated following program guidance.')  # type: Optional[str]
    Description = StringDescriptor(
        'Description', _required,
        docstring='A description of the labeling source')  # type: Optional[str]

    def __init__(self, SourceType='Unknown', SourceID=None, Description=None, **kwargs):
        """
        Parameters
        ----------
        SourceType : str
        SourceID : None|str
        Description : None|str
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        self.SourceType = SourceType
        self.SourceID = SourceID
        self.Description = Description
        super(LabelSourceType, self).__init__(**kwargs)



"""
Module for reading and writing CRSD version 1.0 files
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valyrie")


import logging
import os
from typing import Union, Tuple, List, Sequence, Dict, BinaryIO, Optional
from collections import OrderedDict

import numpy

from sarpy.io.general.utils import is_file_like
from sarpy.io.general.base import BaseReader, SarpyIOError
from sarpy.io.general.data_segment import DataSegment, NumpyMemmapSegment
from sarpy.io.general.slice_parsing import verify_subscript, verify_slice

from sarpy.io.phase_history.cphd import CPHDWritingDetails, CPHDWriter1, \
    AmpScalingFunction

from sarpy.io.received.crsd1_elements.CRSD import CRSDType, CRSDHeader, \
    CRSD_SECTION_TERMINATOR
from sarpy.io.received.base import CRSDTypeReader
from sarpy.io.received.crsd_schema import get_namespace, get_default_tuple

logger = logging.getLogger(__name__)

_unhandled_version_text = 'Got unhandled CRSD version number `{}`'
_missing_channel_identifier_text = 'Cannot find CRSD channel for identifier `{}`'
_index_range_text = 'index must be in the range `[0, {})`'


#########
# Object for parsing CRSD elements

class CRSDDetails(object):
    """
    The basic CRSD element parser.
    """

    __slots__ = (
        '_file_name', '_file_object', '_close_after', '_crsd_version', '_crsd_header', '_crsd_meta')

    def __init__(self, file_object: Union[str, BinaryIO]):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
            The path to or file like object referencing the CRSD file.
        """

        self._crsd_version = None
        self._crsd_header = None
        self._crsd_meta = None
        self._close_after = False

        if isinstance(file_object, str):
            if not os.path.exists(file_object) or not os.path.isfile(file_object):
                raise SarpyIOError('path {} does not exist or is not a file'.format(file_object))
            self._file_name = file_object
            self._file_object = open(file_object, 'rb')
            self._close_after = True
        elif is_file_like(file_object):
            self._file_object = file_object
            if hasattr(file_object, 'name') and isinstance(file_object.name, str):
                self._file_name = file_object.name
            else:
                self._file_name = '<file like object>'
            self._close_after = False
        else:
            raise TypeError('Got unsupported input type {}'.format(type(file_object)))

        self._file_object.seek(0, os.SEEK_SET)
        head_bytes = self._file_object.read(10)
        if not isinstance(head_bytes, bytes):
            raise ValueError('Input file like object not open in bytes mode.')
        if not head_bytes.startswith(b'CRSD'):
            raise SarpyIOError('File {} does not appear to be a CRSD file.'.format(self.file_name))

        self._extract_version()
        self._extract_header()
        self._extract_crsd()

    @property
    def file_name(self) -> str:
        """
        str: The CRSD filename.
        """

        return self._file_name

    @property
    def file_object(self) -> BinaryIO:
        """
        BinaryIO: The binary file object
        """

        return self._file_object

    @property
    def crsd_version(self) -> str:
        """
        str: The CRSD version.
        """

        return self._crsd_version

    @property
    def crsd_header(self) -> CRSDHeader:
        """
        CRSDHeader: The CRSD header object
        """

        return self._crsd_header

    @property
    def crsd_meta(self) -> CRSDType:
        """
        CRSDType: The CRSD structure, which is version dependent.
        """

        return self._crsd_meta

    def _extract_version(self) -> None:
        """
        Extract the version number from the file. This will advance the file
        object to the end of the initial header line.
        """

        self._file_object.seek(0, os.SEEK_SET)
        head_line = self._file_object.readline().strip()
        parts = head_line.split(b'/')
        if len(parts) != 2:
            raise ValueError('Cannot extract CRSD version number from line {}'.format(head_line))
        if parts[0] != b'CRSD':
            raise ValueError('"{}" does not conform to a CRSD file type header'.format(head_line))
        crsd_version = parts[1].strip().decode('utf-8')
        self._crsd_version = crsd_version

    def _extract_header(self) -> None:
        """
        Extract the header from the file. The file object is assumed to be advanced
        to the header location. This will advance to the file object to the end of
        the header section.
        """

        if self.crsd_version.startswith('1.'):
            self._crsd_header = CRSDHeader.from_file_object(self._file_object)
        else:
            raise ValueError(_unhandled_version_text.format(self.crsd_version))

    def _extract_crsd(self) -> None:
        """
        Extract and interpret the CRSD structure from the file.
        """

        xml = self.get_crsd_bytes()
        if self.crsd_version.startswith('1.'):
            the_type = CRSDType
        else:
            raise ValueError(_unhandled_version_text.format(self.crsd_version))

        self._crsd_meta = the_type.from_xml_string(xml)

    def get_crsd_bytes(self) -> bytes:
        """
        Extract the (uninterpreted) bytes representation of the CRSD structure.

        Returns
        -------
        bytes
        """

        header = self.crsd_header
        if header is None:
            raise ValueError('No crsd_header populated.')

        if self.crsd_version.startswith('1.'):
            assert isinstance(header, CRSDHeader)
            # extract the xml data
            self._file_object.seek(header.XML_BLOCK_BYTE_OFFSET, os.SEEK_SET)
            xml = self._file_object.read(header.XML_BLOCK_SIZE)
        else:
            raise ValueError(_unhandled_version_text.format(self.crsd_version))
        return xml

    def __del__(self):
        if self._close_after:
            self._close_after = False
            # noinspection PyBroadException
            try:
                self._file_object.close()
            except Exception:
                pass


def _validate_crsd_details(
        crsd_details: Union[str, CRSDDetails],
        version: Union[None, str, Sequence[str]] = None) -> CRSDDetails:
    """
    Validate the input argument.

    Parameters
    ----------
    crsd_details : str|CRSDDetails
    version : None|str|Sequence[str]

    Returns
    -------
    CRSDDetails
    """

    if isinstance(crsd_details, str):
        crsd_details = CRSDDetails(crsd_details)

    if not isinstance(crsd_details, CRSDDetails):
        raise TypeError('crsd_details is required to be a file path to a CRSD file '
                        'or CRSDDetails, got type {}'.format(crsd_details))

    if version is not None:
        if isinstance(version, str) and not crsd_details.crsd_version.startswith(version):
            raise ValueError(
                'This CRSD file is required to be version {},\n\t'
                'got {}'.format(version, crsd_details.crsd_version))
        else:
            val = False
            for entry in version:
                if crsd_details.crsd_version.startswith(entry):
                    val = True
                    break
            if not val:
                raise ValueError(
                    'This CRSD file is required to be one of version {},\n\t'
                    'got {}'.format(version, crsd_details.crsd_version))

    return crsd_details


class CRSDReader(CRSDTypeReader):
    """
    The Abstract CRSD reader instance, which just selects the proper CRSD reader
    class based on the CRSD version. Note that there is no __init__ method for
    this class, and it would be skipped regardless. Ensure that you make a direct
    call to the BaseReader.__init__() method when extending this class.

    **Updated in version 1.3.0** for reading changes.
    """

    __slots__ = ('_crsd_details', )

    def __new__(cls, *args, **kwargs):
        if len(args) == 0:
            raise ValueError(
                'The first argument of the constructor is required to be a file_path '
                'or CRSDDetails instance.')
        if is_file_like(args[0]):
            raise ValueError('File like object input not supported for CRSD reading at this time.')
        crsd_details = _validate_crsd_details(args[0])

        if crsd_details.crsd_version.startswith('1.'):
            return object.__new__(CRSDReader1)
        else:
            raise ValueError('Got unhandled CRSD version {}'.format(crsd_details.crsd_version))

    @property
    def crsd_details(self) -> CRSDDetails:
        """
        CRSDDetails: The crsd details object.
        """

        return self._crsd_details

    @property
    def crsd_version(self) -> str:
        """
        str: The CRSD version.
        """

        return self.crsd_details.crsd_version

    @property
    def crsd_header(self) -> CRSDHeader:
        """
        CRSDHeader: The CRSD header object
        """

        return self.crsd_details.crsd_header

    @property
    def file_name(self) -> str:
        return self.crsd_details.file_name

    def read_support_array(self,
                           index: Union[int, str],
                           *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]]) -> numpy.ndarray:
        raise NotImplementedError

    def read_support_block(self) -> Dict[str, numpy.ndarray]:
        raise NotImplementedError

    def read_pvp_variable(
            self,
            variable: str,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> Optional[numpy.ndarray]:
        raise NotImplementedError

    def read_pvp_array(
            self,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> numpy.ndarray:
        raise NotImplementedError

    def read_pvp_block(self) -> Dict[str, numpy.ndarray]:
        raise NotImplementedError

    def read_signal_block(self) -> Dict[str, numpy.ndarray]:
        raise NotImplementedError

    def read_signal_block_raw(self) -> Dict[str, numpy.ndarray]:
        raise NotImplementedError

    def close(self):
        CRSDTypeReader.close(self)
        if hasattr(self, '_crsd_details'):
            if hasattr(self._crsd_details, 'close'):
                self._crsd_details.close()
            del self._crsd_details


class CRSDReader1(CRSDReader):
    """
    The CRSD version 1 reader.

    **Updated in version 1.3.0** for reading changes.
    """
    _allowed_versions = ('1.0', )

    def __new__(cls, *args, **kwargs):
        # we must override here, to avoid recursion with
        # the CRSDReader parent
        return object.__new__(cls)

    def __init__(
            self,
            crsd_details: Union[str, CRSDDetails]):
        """

        Parameters
        ----------
        crsd_details : str|CRSDDetails
        """

        self._channel_map = None  # type: Union[None, Dict[str, int]]
        self._support_array_map = None  # type: Union[None, Dict[str, int]]
        self._pvp_memmap = None  # type: Union[None, Dict[str, numpy.ndarray]]
        self._support_array_memmap = None  # type: Union[None, Dict[str, numpy.ndarray]]
        self._crsd_details = _validate_crsd_details(crsd_details, version=self._allowed_versions)

        CRSDTypeReader.__init__(self, None, self._crsd_details.crsd_meta)
        # set data segments after setting up the pvp information, because
        #   we need the AmpSf to set up the format function for the data segment
        self._create_pvp_memmaps()
        self._create_support_array_memmaps()

        data_segments = self._create_data_segments()
        BaseReader.__init__(self, data_segments, reader_type='CRSD')

    @property
    def crsd_meta(self) -> CRSDType:
        """
        CRSDType: the crsd meta_data.
        """

        return self._crsd_meta

    @property
    def crsd_header(self) -> CRSDHeader:
        """
        CRSDHeader: The CRSD header object.
        """

        return self.crsd_details.crsd_header

    def _create_data_segments(self) -> List[DataSegment]:
        """
        Helper method for creating the various signal data segments.

        Returns
        -------
        List[DataSegment]
        """

        data_segments = []

        data = self.crsd_meta.Data
        sample_type = data.SignalArrayFormat

        if sample_type == "CF8":
            raw_dtype = numpy.dtype('>f4')
        elif sample_type == "CI4":
            raw_dtype = numpy.dtype('>i2')
        elif sample_type == "CI2":
            raw_dtype = numpy.dtype('>i1')
        else:
            raise ValueError('Got unhandled signal array format {}'.format(sample_type))

        block_offset = self.crsd_header.SIGNAL_BLOCK_BYTE_OFFSET
        for entry in data.Channels:
            amp_sf = self.read_pvp_variable('AmpSF', entry.Identifier)
            format_function = AmpScalingFunction(raw_dtype, amplitude_scaling=amp_sf)
            raw_shape = (entry.NumVectors, entry.NumSamples, 2)
            data_offset = entry.SignalArrayByteOffset
            data_segments.append(
                NumpyMemmapSegment(
                    self.crsd_details.file_object, block_offset+data_offset,
                    raw_dtype, raw_shape, formatted_dtype='complex64', formatted_shape=raw_shape[:2],
                    format_function=format_function, close_file=False))
        return data_segments

    def _create_pvp_memmaps(self) -> None:
        """
        Helper method which creates the pvp mem_maps.

        Returns
        -------
        None
        """

        self._pvp_memmap = None
        if self.crsd_meta.Data.Channels is None:
            logger.error('No Data.Channels defined.')
            return
        if self.crsd_meta.PVP is None:
            logger.error('No PVP object defined.')
            return

        pvp_dtype = self.crsd_meta.PVP.get_vector_dtype()
        self._pvp_memmap = OrderedDict()
        self._channel_map = OrderedDict()
        for i, entry in enumerate(self.crsd_meta.Data.Channels):
            self._channel_map[entry.Identifier] = i
            offset = self.crsd_header.PVP_BLOCK_BYTE_OFFSET + entry.PVPArrayByteOffset
            shape = (entry.NumVectors, )
            self._pvp_memmap[entry.Identifier] = numpy.memmap(
                self.crsd_details.file_name, dtype=pvp_dtype, mode='r', offset=offset, shape=shape)

    def _create_support_array_memmaps(self) -> None:
        """
        Helper method which creates the support array mem_maps.

        Returns
        -------
        None
        """

        if self.crsd_meta.Data.SupportArrays is None:
            self._support_array_memmap = None
            return

        self._support_array_memmap = OrderedDict()
        for i, entry in enumerate(self.crsd_meta.Data.SupportArrays):
            # extract the support array metadata details
            details = self.crsd_meta.SupportArray.find_support_array(entry.Identifier)
            # determine array byte offset
            offset = self.crsd_header.SUPPORT_BLOCK_BYTE_OFFSET + entry.ArrayByteOffset
            # determine numpy dtype and depth of array
            dtype, depth = details.get_numpy_format()
            # set up the numpy memory map
            shape = (entry.NumRows, entry.NumCols) if depth == 1 else (entry.NumRows, entry.NumCols, depth)
            self._support_array_memmap[entry.Identifier] = numpy.memmap(
                self.crsd_details.file_name, dtype=dtype, mode='r', offset=offset, shape=shape)

    def _validate_index(self, index: Union[int, str]) -> int:
        """
        Get corresponding integer index for CRSD channel.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        int
        """

        crsd_meta = self.crsd_details.crsd_meta

        if isinstance(index, str):
            if index in self._channel_map:
                return self._channel_map[index]
            else:
                raise KeyError(_missing_channel_identifier_text.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < crsd_meta.Data.NumCRSDChannels):
                raise ValueError(_index_range_text.format(crsd_meta.Data.NumCRSDChannels))
            return int_index

    def _validate_index_key(self, index: Union[int, str]) -> str:
        """
        Gets the corresponding identifier for the CPHD channel.

        Parameters
        ----------
        index : int|str

        Returns
        -------
        str
        """

        crsd_meta = self.crsd_details.crsd_meta

        if isinstance(index, str):
            if index in self._channel_map:
                return index
            else:
                raise KeyError(_missing_channel_identifier_text.format(index))
        else:
            int_index = int(index)
            if not (0 <= int_index < crsd_meta.Data.NumCRSDChannels):
                raise ValueError(_index_range_text.format(crsd_meta.Data.NumCRSDChannels))
            return crsd_meta.Data.Channels[int_index].Identifier

    def read_support_array(
            self,
            index: Union[int, str],
            *ranges) -> numpy.ndarray:
        # find the support array identifier
        if isinstance(index, int):
            the_entry = self.crsd_meta.Data.SupportArrays[index]
            index = the_entry.Identifier
        if not isinstance(index, str):
            raise TypeError('Got unexpected type {} for identifier'.format(type(index)))

        the_memmap = self._support_array_memmap[index]

        if len(ranges) == 0:
            return numpy.copy(the_memmap[:])

        # noinspection PyTypeChecker
        subscript = verify_subscript(ranges, the_memmap.shape)
        return numpy.copy(the_memmap[subscript])

    def read_support_block(self) -> Dict:
        if self.crsd_meta.Data.SupportArrays:
            return {
                sa.Identifier: self.read_support_array(sa.Identifier)
                for sa in self.crsd_meta.Data.SupportArrays}
        else:
            return {}

    def read_pvp_variable(self, variable, index, the_range=None):
        index_key = self._validate_index_key(index)
        the_memmap = self._pvp_memmap[index_key]
        the_slice = verify_slice(the_range, the_memmap.shape[0])
        if variable in the_memmap.dtype.fields:
            return numpy.copy(the_memmap[variable][the_slice])
        else:
            return None

    def read_pvp_array(self, index, the_range=None):
        index_key = self._validate_index_key(index)
        the_memmap = self._pvp_memmap[index_key]
        the_slice = verify_slice(the_range, the_memmap.shape[0])
        return numpy.copy(the_memmap[the_slice])

    def read_pvp_block(self) -> Dict[str, numpy.ndarray]:
        return {chan.Identifier: self.read_pvp_array(chan.Identifier) for chan in self.crsd_meta.Data.Channels}

    def read_signal_block(self) -> Dict[str, numpy.ndarray]:
        return {chan.Identifier: numpy.copy(self.read(index=chan.Identifier)) for chan in self.crsd_meta.Data.Channels}

    def read_signal_block_raw(self) -> Dict[str, numpy.ndarray]:
        return {chan.Identifier: numpy.copy(self.read_raw(index=chan.Identifier)) for chan in self.crsd_meta.Data.Channels}

    def read_chip(self,
             *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
             index: Union[int, str] = 0,
             squeeze: bool = True) -> numpy.ndarray:
        """
        This is identical to :meth:`read`, and presented for backwards compatibility.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
        index : int|str
        squeeze : bool

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        :meth:`read`.
        """

        return self.__call__(*ranges, index=index, raw=False, squeeze=squeeze)

    def read(self,
             *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
             index: Union[int, str] = 0,
             squeeze: bool = True) -> numpy.ndarray:
        """
        Read formatted data from the given data segment. Note this is an alias to the
        :meth:`__call__` called as
        :code:`reader(*ranges, index=index, raw=False, squeeze=squeeze)`.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
            The slice definition appropriate for `data_segment[index].read()` usage.
        index : int|str
            The data_segment index or channel identifier.
        squeeze : bool
            Squeeze length 1 dimensions out of the shape of the return array?

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.read`.
        """

        return self.__call__(*ranges, index=index, raw=False, squeeze=squeeze)

    def read_raw(self,
                 *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
                 index: Union[int, str] = 0,
                 squeeze: bool = True) -> numpy.ndarray:
        """
        Read raw data from the given data segment. Note this is an alias to the
        :meth:`__call__` called as
        :code:`reader(*ranges, index=index, raw=True, squeeze=squeeze)`.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
            The slice definition appropriate for `data_segment[index].read()` usage.
        index : int|str
            The data_segment index or crsd channel identifier.
        squeeze : bool
            Squeeze length 1 dimensions out of the shape of the return array?

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.read_raw`.
        """

        return self.__call__(*ranges, index=index, raw=True, squeeze=squeeze)

    def __call__(self,
                 *ranges: Sequence[Union[None, int, slice]],
                 index: int = 0,
                 raw: bool = False,
                 squeeze: bool = True) -> numpy.ndarray:
        index = self._validate_index(index)
        return BaseReader.__call__(self, *ranges, index=index, raw=raw, squeeze=squeeze)


def is_a(file_name: str) -> Optional[CRSDReader]:
    """
    Tests whether a given file_name corresponds to a CRSD file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    CRSDReader1|None
        Appropriate `CRSDReader` instance if CRSD file, `None` otherwise
    """

    try:
        crsd_details = CRSDDetails(file_name)
        logger.info('File {} is determined to be a CRSD version {} file.'.format(file_name, crsd_details.crsd_version))
        return CRSDReader(crsd_details)
    except SarpyIOError:
        # we don't want to catch parsing errors, for now?
        return None


###########
# writer

class CRSDWritingDetails(CPHDWritingDetails):

    @property
    def header(self) -> CRSDHeader:
        return self._header

    def _set_header(self, check_older_version: bool):
        if check_older_version:
            use_version_tuple = self.meta.version_required()
        else:
            use_version_tuple = get_default_tuple()
        use_version_string = '{}.{}.{}'.format(*use_version_tuple)
        self._header = self.meta.make_file_header(use_version=use_version_string)

    @property
    def meta(self) -> CRSDType:
        """
        CPSDType: The metadata
        """

        return self._meta

    @meta.setter
    def meta(self, value: CRSDType):
        if self._meta is not None:
            raise ValueError('meta is read only once initialized.')
        if not isinstance(value, CRSDType):
            raise TypeError('meta must be of type {}'.format(CRSDType))
        self._meta = value

    def write_header(
            self,
            file_object: BinaryIO,
            overwrite: bool = False) -> None:
        """
        Write the header.The file object will be advanced to the end of the
        block, if writing occurs.

        Parameters
        ----------
        file_object : BinaryIO
        overwrite : bool
            Overwrite, if previously written?

        Returns
        -------
        None
        """

        if self._header_written and not overwrite:
            return

        file_object.write(self.header.to_string().encode())
        file_object.write(CRSD_SECTION_TERMINATOR)
        # write xml
        file_object.seek(self.header.XML_BLOCK_BYTE_OFFSET, os.SEEK_SET)
        file_object.write(self.meta.to_xml_bytes(urn=get_namespace(self.use_version)))
        file_object.write(CRSD_SECTION_TERMINATOR)
        self._header_written = True


class CRSDWriter1(CPHDWriter1):
    """
    The CRSD version 1 writer.

    **Updated in version 1.3.0** for writing changes.
    """
    _writing_details_type = CRSDWritingDetails

    def __init__(
            self,
            file_object: Union[str, BinaryIO],
            meta: Optional[CRSDType] = None,
            writing_details: Optional[CRSDWritingDetails] = None,
            check_existence: bool = True):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
        meta : None|CRSDType
        writing_details : None|CRSDWritingDetails
        check_existence : bool
            Should we check if the given file already exists, and raises an exception if so?
        """

        CPHDWriter1.__init__(
            self, file_object, meta=meta, writing_details=writing_details,
            check_existence=check_existence)

    @property
    def writing_details(self) -> CRSDWritingDetails:
        return self._writing_details

    @writing_details.setter
    def writing_details(self, value: CRSDWritingDetails):
        if self._writing_details is not None:
            raise ValueError('writing_details is read-only')
        if not isinstance(value, CRSDWritingDetails):
            raise TypeError('writing_details must be of type {}'.format(CRSDWritingDetails))
        self._writing_details = value

    @property
    def file_name(self) -> Optional[str]:
        return self._file_name

    @property
    def meta(self) -> CRSDType:
        """
        CRSDType: The metadata
        """

        return self.writing_details.meta





from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class GEOPSBType(TREElement):
    def __init__(self, value):
        super(GEOPSBType, self).__init__()
        self.add_field('TYP', 's', 3, value)
        self.add_field('UNI', 's', 3, value)
        self.add_field('DAG', 's', 80, value)
        self.add_field('DCD', 's', 4, value)
        self.add_field('ELL', 's', 80, value)
        self.add_field('ELC', 's', 3, value)
        self.add_field('DVR', 's', 80, value)
        self.add_field('VDCDVR', 's', 4, value)
        self.add_field('SDA', 's', 80, value)
        self.add_field('VDCSDA', 's', 4, value)
        self.add_field('ZOR', 's', 15, value)
        self.add_field('GRD', 's', 3, value)
        self.add_field('GRN', 's', 80, value)
        self.add_field('ZNA', 's', 4, value)


class GEOPSB(TREExtension):
    _tag_value = 'GEOPSB'
    _data_type = GEOPSBType



"""
Common ortho-rectification elements
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import os
from typing import Union, Tuple, List, Optional, Sequence

import numpy

from sarpy.io.complex.converter import open_complex
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.general.slice_parsing import verify_subscript
from sarpy.io.complex.utils import get_fetch_block_size, extract_blocks

from sarpy.geometry.geocoords import ecf_to_geodetic
from sarpy.visualization.remap import RemapFunction

from .ortho_methods import OrthorectificationHelper

logger = logging.getLogger(__name__)


class FullResolutionFetcher(object):
    """
    This is a base class for provided a simple API for processing schemes where
    full resolution is required along the processing dimension, so sub-sampling
    along the processing dimension does not decrease the amount of data which
    must be fetched.
    """

    __slots__ = (
        '_reader', '_index', '_sicd', '_dimension', '_data_size', '_block_size')

    def __init__(
            self,
            reader: Union[str, SICDTypeReader],
            dimension: int = 0,
            index: int = 0,
            block_size: Union[None, int, float] = 10):
        """

        Parameters
        ----------
        reader : str|SICDTypeReader
            Input file path or reader object, which must be of sicd type.
        dimension : int
            The dimension over which to split the sub-aperture.
        index : int
            The sicd index to use.
        block_size : None|int|float
            The approximate processing block size to fetch, given in MB. The
            minimum value for use here will be 0.25. `None` represents processing
            as a single block.
        """

        self._index = None  # set explicitly
        self._sicd = None  # set with index setter
        self._dimension = None  # set explicitly
        self._data_size = None  # set with index setter
        self._block_size = None  # set explicitly

        # validate the reader
        if isinstance(reader, str):
            reader = open_complex(reader)
        if not isinstance(reader, SICDTypeReader):
            raise TypeError('reader is required to be a path name for a sicd-type image, '
                            'or an instance of a reader object.')
        self._reader = reader
        # set the other properties
        self.dimension = dimension
        self.index = index
        self.block_size = block_size

    @property
    def reader(self) -> SICDTypeReader:
        """
        SICDTypeReader: The reader instance.
        """

        return self._reader

    @property
    def dimension(self) -> int:
        """
        int: The dimension along which to perform the color subaperture split.
        """

        return self._dimension

    @dimension.setter
    def dimension(self, value):
        value = int(value)
        if value not in [0, 1]:
            raise ValueError('dimension must be 0 or 1, got {}'.format(value))
        self._dimension = value

    @property
    def data_size(self) -> Tuple[int, ...]:
        """
        Tuple[int, ...]: The data size for the reader at the given index.
        """

        return self._data_size

    @property
    def index(self) -> int:
        """
        int: The index of the reader.
        """

        return self._index

    @index.setter
    def index(self, value):
        self._set_index(value)

    def _set_index(self, value):
        value = int(value)
        if value < 0:
            raise ValueError('The index must be a non-negative integer, got {}'.format(value))

        sicds = self.reader.get_sicds_as_tuple()
        if value >= len(sicds):
            raise ValueError('The index must be less than the sicd count.')
        self._index = value
        self._sicd = sicds[value]
        self._data_size = self.reader.get_data_size_as_tuple()[value]

    @property
    def block_size(self) -> Optional[float]:
        """
        None|float: The approximate processing block size in MB, where `None`
        represents processing in a single block.
        """

        return self._block_size

    @block_size.setter
    def block_size(self, value):
        if value is None:
            self._block_size = None
        else:
            value = float(value)
            if value < 0.25:
                value = 0.25
            self._block_size = value

    @property
    def block_size_in_bytes(self) -> Optional[int]:
        """
        None|int: The approximate processing block size in bytes.
        """

        return None if self._block_size is None else int(self._block_size*(2**20))

    @property
    def sicd(self) -> SICDType:
        """
        SICDType: The sicd structure.
        """

        return self._sicd

    def _parse_slicing(
            self,
            item: Union[None, int, slice, Tuple[Union[int, slice], ...]]) -> Tuple[slice, slice, Optional[int]]:
        if isinstance(item, tuple) and len(item) > 2:
            if len(item) > 3:
                raise ValueError('Got unexpected subscript {}'.format(item))
            if len(item) == 3:
                if not isinstance(item[2], int):
                    raise ValueError('Got unexpected subscript {}'.format(item))
                return verify_subscript(item[:2], self._data_size) + (item[2], )
        return verify_subscript(item, self._data_size) + (None, )

    def get_fetch_block_size(self, start_element: int, stop_element: int) -> int:
        """
        Gets the fetch block size for the given full resolution section.
        This assumes that the fetched data will be 8 bytes per pixel, in
        accordance with single band complex64 data.

        Parameters
        ----------
        start_element : int
        stop_element : int

        Returns
        -------
        int
        """

        return get_fetch_block_size(start_element, stop_element, self.block_size_in_bytes, bands=1)

    @staticmethod
    def extract_blocks(
            the_range: Union[slice, Tuple[int, int, int]],
            index_block_size: Union[None, int, float]) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int]]]:
        """
        Convert the single range definition into a series of range definitions in
        keeping with fetching of the appropriate block sizes.

        Parameters
        ----------
        the_range : slice|Tuple[int, int, int]
            The input (off processing axis) range.
        index_block_size : None|int|float
            The size of blocks (number of indices).

        Returns
        -------
        range_definitions: List[Tuple[int, int, int]]
            The sequence of range definitions `(start index, stop index, step)`
            relative to the overall image.
        limit_indices: List[Tuple[int, int]]
            The sequence of start/stop indices for positioning of the given
            range relative to the original range.
        """

        if isinstance(the_range, slice):
            if the_range.stop is None:
                if the_range.step > 0:
                    raise ValueError('Got unexpected slice {}'.format(the_range))
                use_range = (the_range.start, -1, the_range.step)
            else:
                use_range = (the_range.start, the_range.stop, the_range.step)
        else:
            use_range = the_range
        # noinspection PyTypeChecker
        return extract_blocks(use_range, index_block_size)

    def _full_row_resolution(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]]) -> numpy.ndarray:
        """
        Perform the full row resolution data, with any appropriate calculations.

        Parameters
        ----------
        row_range : slice|Tuple[int, int, int]
        col_range : slice|Tuple[int, int, int]

        Returns
        -------
        numpy.ndarray
        """

        if isinstance(row_range, Sequence):
            row_range = slice(*row_range)
        if isinstance(col_range, Sequence):
            col_range = slice(*col_range)

        # fetch the data and perform the csi calculation
        if row_range.step not in [1, -1]:
            raise ValueError('The step for row_range must be +/- 1, for full row resolution data.')
        data = self.reader[(row_range, col_range, self.index)]

        if data.ndim < 2:
            data = numpy.reshape(data, (-1, 1))
        # handle nonsense data with zeros
        data[~numpy.isfinite(data)] = 0
        return data

    def _full_column_resolution(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]]) -> numpy.ndarray:
        """
        Perform the full column resolution data, with any appropriate calculations.

        Parameters
        ----------
        row_range : Tuple[int, int, int]
        col_range : Tuple[int, int, int]

        Returns
        -------
        numpy.ndarray
        """

        if isinstance(row_range, Sequence):
            row_range = slice(*row_range)
        if isinstance(col_range, Sequence):
            col_range = slice(*col_range)

        # fetch the data and perform the csi calculation
        if col_range.step not in [1, -1]:
            raise ValueError('The step for col_range must be +/- 1, for full col resolution data.')
        data = self.reader[(row_range, col_range, self.index)]

        if data.ndim < 2:
            data = numpy.reshape(data, (1, -1))
        # handle nonsense data with zeros
        data[~numpy.isfinite(data)] = 0
        return data

    def _prepare_output(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]]) -> numpy.ndarray:
        """
        Prepare the output workspace for :func:`__getitem__`.

        Parameters
        ----------
        row_range : slice|Tuple[int, int, int]
        col_range : slice|Tuple[int, int, int]

        Returns
        -------
        numpy.ndarray
        """

        row_count = int((row_range[1] - row_range[0]) / float(row_range[2]))
        col_count = int((col_range[1] - col_range[0]) / float(col_range[2]))
        out_size = (row_count, col_count)
        return numpy.zeros(out_size, dtype=numpy.complex64)

    def __getitem__(self, subscript) -> numpy.ndarray:
        """
        Fetches the processed data based on the input slice.

        Parameters
        ----------
        subscript

        Returns
        -------
        numpy.ndarray
        """

        subscript = verify_subscript(subscript, self.data_size)
        return self.reader.read(*subscript, index=self.index)


class OrthorectificationIterator(object):
    """
    This provides a generator for an Orthorectification process on a given
    reader/index/(pixel) bounds.
    """

    __slots__ = (
        '_calculator', '_ortho_helper', '_pixel_bounds', '_ortho_bounds',
        '_this_index', '_iteration_blocks', '_remap_function')

    def __init__(
            self,
            ortho_helper: OrthorectificationHelper,
            calculator: Optional[FullResolutionFetcher] = None,
            bounds: Union[None, numpy.ndarray, tuple, list] = None,
            remap_function: Optional[RemapFunction] = None,
            recalc_remap_globals: bool = False):
        """

        Parameters
        ----------
        ortho_helper : OrthorectificationHelper
            The ortho-rectification helper.
        calculator : None|FullResolutionFetcher
            The FullResolutionFetcher instance. If not provided, then this will
            default to a base FullResolutionFetcher instance - which is only
            useful for a basic detected image.
        bounds : None|numpy.ndarray|list|tuple
            The pixel bounds of the form `(min row, max row, min col, max col)`.
            This will default to the full image.
        remap_function : None|RemapFunction
            The remap function to apply, if desired.
        recalc_remap_globals : bool
            Only applies if a remap function is provided, should we recalculate
            any required global parameters? This will automatically happen if
            they are not already set.
        """

        self._this_index = None
        self._iteration_blocks = None
        self._remap_function = None

        # validate ortho_helper
        if not isinstance(ortho_helper, OrthorectificationHelper):
            raise TypeError(
                'ortho_helper must be an instance of OrthorectificationHelper, got '
                'type {}'.format(type(ortho_helper)))
        self._ortho_helper = ortho_helper

        # validate calculator
        if calculator is None:
            calculator = FullResolutionFetcher(ortho_helper.reader, index=ortho_helper.index, dimension=0)
        if not isinstance(calculator, FullResolutionFetcher):
            raise TypeError(
                'calculator must be an instance of FullResolutionFetcher, got '
                'type {}'.format(type(calculator)))
        self._calculator = calculator

        if ortho_helper.reader.file_name is not None and calculator.reader.file_name is not None and \
                os.path.abspath(ortho_helper.reader.file_name) != os.path.abspath(calculator.reader.file_name):
            raise ValueError(
                'ortho_helper has reader for file {}, while calculator has reader '
                'for file {}'.format(ortho_helper.reader.file_name, calculator.reader.file_name))
        if ortho_helper.index != calculator.index:
            raise ValueError(
                'ortho_helper is using index {}, while calculator is using '
                'index {}'.format(ortho_helper.index, calculator.index))

        # validate the bounds
        if bounds is not None:
            pixel_bounds, pixel_rectangle = ortho_helper.bounds_to_rectangle(bounds)
            # get the corresponding ortho bounds
            ortho_bounds = ortho_helper.get_orthorectification_bounds_from_pixel_object(pixel_rectangle)
        else:
            ortho_bounds = ortho_helper.get_full_ortho_bounds()
            ortho_bounds, nominal_pixel_bounds = ortho_helper.extract_pixel_bounds(ortho_bounds)
            # extract the values - ensure that things are within proper image bounds
            pixel_bounds = ortho_helper.get_real_pixel_bounds(nominal_pixel_bounds)

        # validate remap function
        if remap_function is None or isinstance(remap_function, RemapFunction):
            self._remap_function = remap_function
        else:
            raise TypeError(
                'remap_function is expected to be an instance of RemapFunction, '
                'got type `{}`'.format(type(remap_function)))

        self._pixel_bounds = pixel_bounds
        self._ortho_bounds = ortho_bounds
        self._prepare_state(recalc_remap_globals=recalc_remap_globals)

    @property
    def ortho_helper(self) -> OrthorectificationHelper:
        """
        OrthorectificationHelper: The ortho-rectification helper.
        """

        return self._ortho_helper

    @property
    def calculator(self) -> FullResolutionFetcher:
        """
        FullResolutionFetcher : The calculator instance.
        """

        return self._calculator

    @property
    def sicd(self) -> SICDType:
        """
        SICDType: The sicd structure.
        """

        return self.calculator.sicd

    @property
    def pixel_bounds(self) -> numpy.ndarray:
        """
        numpy.ndarray : Of the form `(row min, row max, col min, col max)`.
        """

        return self._pixel_bounds

    @property
    def ortho_bounds(self) -> numpy.ndarray:
        """
        numpy.ndarray : Of the form `(row min, row max, col min, col max)`. Note that
        these are "unnormalized" orthorectified pixel coordinates.
        """

        return self._ortho_bounds

    @property
    def ortho_data_size(self) -> Tuple[int, int]:
        """
        Tuple[int, int] : The size of the overall ortho-rectified output.
        """

        return (
            int(self.ortho_bounds[1] - self.ortho_bounds[0]),
            int(self.ortho_bounds[3] - self.ortho_bounds[2]))

    @property
    def remap_function(self) -> Optional[RemapFunction]:
        """
        None|RemapFunction: The remap function to be applied.
        """

        return self._remap_function

    def get_ecf_image_corners(self) -> Optional[numpy.ndarray]:
        """
        The corner points of the overall ortho-rectified output in ECF
        coordinates. The ordering of these points follows the SICD convention.

        Returns
        -------
        numpy.ndarray
        """

        if self.ortho_bounds is None:
            return None
        _, ortho_pixel_corners = self._ortho_helper.bounds_to_rectangle(self.ortho_bounds)
        return self._ortho_helper.proj_helper.ortho_to_ecf(ortho_pixel_corners)

    def get_llh_image_corners(self) -> Optional[numpy.ndarray]:
        """
        The corner points of the overall ortho-rectified output in Lat/Lon/HAE
        coordinates. The ordering of these points follows the SICD convention.

        Returns
        -------
        None|numpy.ndarray
        """

        ecf_corners = self.get_ecf_image_corners()
        if ecf_corners is None:
            return None
        else:
            return ecf_to_geodetic(ecf_corners)

    def _prepare_state(self, recalc_remap_globals: bool = False) -> None:
        """
        Prepare the iteration state.

        Parameters
        ----------
        recalc_remap_globals : bool

        Returns
        -------
        None
        """

        if self.calculator.dimension == 0:
            column_block_size = self.calculator.get_fetch_block_size(self.ortho_bounds[0], self.ortho_bounds[1])
            self._iteration_blocks, _ = self.calculator.extract_blocks(
                (self.ortho_bounds[2], self.ortho_bounds[3], 1), column_block_size)
        else:
            row_block_size = self.calculator.get_fetch_block_size(self.ortho_bounds[2], self.ortho_bounds[3])
            self._iteration_blocks, _ = self.calculator.extract_blocks(
                (self.ortho_bounds[0], self.ortho_bounds[1], 1), row_block_size)

        if self.remap_function is not None and \
                (recalc_remap_globals or not self.remap_function.are_global_parameters_set):
            self.remap_function.calculate_global_parameters_from_reader(
                self.ortho_helper.reader, index=self.ortho_helper.index, pixel_bounds=self.pixel_bounds)

    @staticmethod
    def _get_ortho_helper(
            pixel_bounds: Union[Tuple[int, int, int, int], numpy.ndarray],
            this_data: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Get helper data for ortho-rectification.

        Parameters
        ----------
        pixel_bounds : Tuple[int, int, int, int]|numpy.ndarray
        this_data : numpy.ndarray

        Returns
        -------
        row_array: numpy.ndarray
        col_array: numpy.ndarray
        """

        rows_temp = pixel_bounds[1] - pixel_bounds[0]
        if this_data.shape[0] == rows_temp:
            row_array = numpy.arange(pixel_bounds[0], pixel_bounds[1])
        elif this_data.shape[0] == (rows_temp - 1):
            row_array = numpy.arange(pixel_bounds[0], pixel_bounds[1] - 1)
        else:
            raise ValueError('Unhandled data size mismatch {} and {}'.format(this_data.shape, rows_temp))

        cols_temp = pixel_bounds[3] - pixel_bounds[2]
        if this_data.shape[1] == cols_temp:
            col_array = numpy.arange(pixel_bounds[2], pixel_bounds[3])
        elif this_data.shape[1] == (cols_temp - 1):
            col_array = numpy.arange(pixel_bounds[2], pixel_bounds[3] - 1)
        else:
            raise ValueError('Unhandled data size mismatch {} and {}'.format(this_data.shape, cols_temp))
        return row_array, col_array

    def _get_orthorectified_version(
            self,
            this_ortho_bounds: numpy.ndarray,
            pixel_bounds: Union[Tuple[int, int, int, int], numpy.ndarray],
            this_data: numpy.ndarray) -> numpy.ndarray:
        """
        Get the orthorectified version from the raw values and pixel information.

        Parameters
        ----------
        this_ortho_bounds
        pixel_bounds
        this_data

        Returns
        -------
        numpy.ndarray
        """

        row_array, col_array = self._get_ortho_helper(pixel_bounds, this_data)
        ortho_data = self._ortho_helper.get_orthorectified_from_array(
            this_ortho_bounds, row_array, col_array, this_data)
        if self.remap_function is None:
            return ortho_data
        else:
            return self.remap_function(ortho_data)

    def _get_state_parameters(
            self,
            pad: int = 10) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Gets the pixel information associated with the current state.

        Parameters
        ----------
        pad : int
            Pad the pixel bounds, to accommodate for any edge cases.

        Returns
        -------
        ortho_bounds: numpy.ndarray
        pixel_bounds: numpy.ndarray
        """

        if self._calculator.dimension == 0:
            this_column_range = self._iteration_blocks[self._this_index]
            # determine the corresponding pixel ranges to encompass these values
            this_ortho_bounds, this_pixel_bounds = self._ortho_helper.extract_pixel_bounds(
                (self.ortho_bounds[0], self.ortho_bounds[1], this_column_range[0], this_column_range[1]))
        else:
            this_row_range = self._iteration_blocks[self._this_index]
            this_ortho_bounds, this_pixel_bounds = self._ortho_helper.extract_pixel_bounds(
                (this_row_range[0], this_row_range[1], self.ortho_bounds[2], self.ortho_bounds[3]))
        
        this_pixel_bounds[0::2] -= pad
        this_pixel_bounds[1::2] += pad
        return this_ortho_bounds, this_pixel_bounds

    def __iter__(self):
        return self

    def __next__(self) -> Tuple[numpy.ndarray, Tuple[int, int]]:
        """
        Get the next iteration of orthorectified data.

        Returns
        -------
        data: numpy.ndarray
        indices: Tuple[int, int]
            The (normalized) indices `(start_row, start_col)` for this section of
            data, relative to overall output shape.
        """

        # NB: this is the Python 3 pattern for iteration
        if self._this_index is None:
            self._this_index = 0
        else:
            self._this_index += 1
        # at this point, _this_index indicates which entry to return
        if self._this_index >= len(self._iteration_blocks):
            self._this_index = None  # reset the iteration scheme
            raise StopIteration()

        this_ortho_bounds, this_pixel_bounds = self._get_state_parameters()
        # accommodate for real pixel limits
        this_pixel_bounds = self._ortho_helper.get_real_pixel_bounds(this_pixel_bounds)
        # extract the csi data and ortho-rectify
        logger.info(
            'Fetching orthorectified coordinate block ({}:{}, {}:{}) of ({}, {})'.format(
                this_ortho_bounds[0] - self.ortho_bounds[0], this_ortho_bounds[1] - self.ortho_bounds[0],
                this_ortho_bounds[2] - self.ortho_bounds[2], this_ortho_bounds[3] - self.ortho_bounds[2],
                self.ortho_bounds[1] - self.ortho_bounds[0], self.ortho_bounds[3] - self.ortho_bounds[2]))
        ortho_data = self._get_orthorectified_version(
            this_ortho_bounds, this_pixel_bounds,
            self._calculator[this_pixel_bounds[0]:this_pixel_bounds[1], this_pixel_bounds[2]:this_pixel_bounds[3]])
        # determine the relative image size
        start_indices = (this_ortho_bounds[0] - self.ortho_bounds[0],
                         this_ortho_bounds[2] - self.ortho_bounds[2])
        return ortho_data, start_indices

    def next(self) -> Tuple[numpy.ndarray, Tuple[int, int]]:
        """
        Get the next iteration of ortho-rectified data.

        Returns
        -------
        data: numpy.ndarray
        indices: Tuple[int, int]
            The (normalized) indices `(start_row, start_col)` for this section of
            data, relative to overall output shape.
        """

        # NB: this is the Python 2 pattern for iteration
        return self.__next__()



"""
The Antenna type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List, Tuple, Optional

import numpy

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.complex.sicd_elements.blocks import Poly1DType, XYZType, \
    XYZPolyType, Poly2DType
from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import FloatDescriptor, StringDescriptor, SerializableDescriptor, \
    BooleanDescriptor, SerializableListDescriptor


class AntCoordFrameType(Serializable):
    """
    Unit vectors that describe the orientation of an Antenna Coordinate Frame
    (ACF) as function of time.
    """

    _fields = ('Identifier', 'XAxisPoly', 'YAxisPoly', 'UseACFPVP')
    _required = ('Identifier', 'XAxisPoly', 'YAxisPoly')
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this ACF.')  # type: str
    XAxisPoly = SerializableDescriptor(
        'XAxisPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna X-Axis unit vector in ECF coordinates as a function '
                  'of time.')  # type: XYZPolyType
    YAxisPoly = SerializableDescriptor(
        'YAxisPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna Y-Axis unit vector in ECF coordinates as a function '
                  'of time.')  # type: XYZPolyType
    UseACFPVP = BooleanDescriptor(
        'UseACFPVP', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: bool

    def __init__(
            self,
            Identifier: str = None,
            XAxisPoly: XYZPolyType = None,
            YAxisPoly: XYZPolyType = None,
            UseACFPVP: Optional[bool] = None,
            **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        XAxisPoly : XYZPolyType
        YAxisPoly : XYZPolyType
        UseACFPVP : None|bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.XAxisPoly = XAxisPoly
        self.YAxisPoly = YAxisPoly
        self.UseACFPVP = UseACFPVP
        super(AntCoordFrameType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.UseACFPVP is not None:
            required = max(required, (1, 1, 0))
        return required


class AntPhaseCenterType(Serializable):
    """
    Parameters that describe each Antenna Phase Center (APC).
    """

    _fields = ('Identifier', 'ACFId', 'APCXYZ')
    _required = _fields
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this APC.')  # type: str
    ACFId = StringDescriptor(
        'ACFId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of Antenna Coordinate Frame used for computing the '
                  'antenna gain and phase patterns.')  # type: str
    APCXYZ = SerializableDescriptor(
        'APCXYZ', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='The APC location in the ACF XYZ coordinate '
                  'frame.')  # type: XYZType

    def __init__(self, Identifier=None, ACFId=None, APCXYZ=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        ACFId : str
        APCXYZ : XYZType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.ACFId = ACFId
        self.APCXYZ = APCXYZ
        super(AntPhaseCenterType, self).__init__(**kwargs)


class GainPhaseArrayType(Serializable):
    """
    Parameters that identify 2-D sampled Gain & Phase patterns at single
    frequency value.
    """

    _fields = ('Freq', 'ArrayId', 'ElementId')
    _required = ('Freq', 'ArrayId')
    _numeric_format = {'Freq': FLOAT_FORMAT}
    # descriptors
    Freq = FloatDescriptor(
        'Freq', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Frequency value for which the sampled Array and Element '
                  'pattern(s) are provided, in Hz.')  # type: float
    ArrayId = StringDescriptor(
        'ArrayId', _required, strict=DEFAULT_STRICT,
        docstring='Support array identifier of the sampled gain/phase of the array '
                  'at ref Frequency.')  # type: str
    ElementId = StringDescriptor(
        'ElementId', _required, strict=DEFAULT_STRICT,
        docstring='Support array identifier of the sampled gain/phase of the element '
                  'at ref frequency.')  # type: str

    def __init__(self, Freq=None, ArrayId=None, ElementId=None, **kwargs):
        """

        Parameters
        ----------
        Freq : float
        ArrayId : str
        ElementId : None|str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Freq = Freq
        self.ArrayId = ArrayId
        self.ElementId = ElementId
        super(GainPhaseArrayType, self).__init__(**kwargs)


class FreqSFType(Serializable, Arrayable):
    _fields = ('DCXSF', 'DCYSF')
    _required = _fields
    _numeric_format = {key: '0.17E' for key in _fields}
    DCXSF = FloatDescriptor(
        'DCXSF', _required, strict=DEFAULT_STRICT, bounds=(0.0, 1.0),
        docstring='')  # type: float
    DCYSF = FloatDescriptor(
        'DCYSF', _required, strict=DEFAULT_STRICT, bounds=(0.0, 1.0),
        docstring='')  # type: float

    def __init__(self, DCXSF=None, DCYSF=None, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DCXSF = DCXSF
        self.DCYSF = DCYSF
        super(FreqSFType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64) -> numpy.ndarray:
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [DCXSF, DCYSF]
        """

        return numpy.array([self.DCXSF, self.DCYSF], dtype=dtype)

    @classmethod
    def from_array(cls, array: numpy.ndarray):
        """
        Construct from an iterable.

        Parameters
        ----------
        array : numpy.ndarray|list|tuple

        Returns
        -------
        FreqSFType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError(
                    'Expected array to be of length 2,\n\t'
                    'and received `{}`'.format(array))
            return cls(DCXSF=array[0], DCYSF=array[1])
        raise ValueError(
            'Expected array to be numpy.ndarray, list, or tuple,\n\t'
            'got `{}`'.format(type(array)))


class AntPolRefType(Serializable, Arrayable):
    """
    Polarization reference type.
    """
    _fields = ('AmpX', 'AmpY', 'PhaseY')
    _required = _fields
    _numeric_format = {key: '0.17E' for key in _fields}
    AmpX = FloatDescriptor(
        'AmpX', _required, strict=DEFAULT_STRICT, bounds=(0.0, 1.0),
        docstring='E-field relative amplitude in ACF X direction')  # type: float
    AmpY = FloatDescriptor(
        'AmpY', _required, strict=DEFAULT_STRICT, bounds=(0.0, 1.0),
        docstring='E-field relative amplitude in ACF Y direction')  # type: float
    PhaseY = FloatDescriptor(
        'PhaseY', _required, strict=DEFAULT_STRICT, bounds=(-0.5, 0.5),
        docstring='Relative phase of the Y E-field '
                  'relative to the X E-field at f=f_0')  # type: float

    def __init__(
            self,
            AmpX: float = None,
            AmpY: float = None,
            PhaseY: float = None,
            **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AmpX = AmpX
        self.AmpY = AmpY
        self.PhaseY = PhaseY
        super(AntPolRefType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64) -> numpy.ndarray:
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [AmpX, AmpY, PhaseY]
        """

        return numpy.array([self.AmpX, self.AmpY, self.PhaseY], dtype=dtype)

    @classmethod
    def from_array(cls, array: numpy.ndarray):
        """
        Construct from an iterable.

        Parameters
        ----------
        array : numpy.ndarray|list|tuple

        Returns
        -------
        AntPolRefType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 3:
                raise ValueError(
                    'Expected array to be of length 3,\n\t'
                    'and received `{}`'.format(array))
            return cls(AmpX=array[0], AmpY=array[1], PhaseY=array[2])
        raise ValueError(
            'Expected array to be numpy.ndarray, list, or tuple,\n\t'
            'got `{}`'.format(type(array)))


class EBType(Serializable):
    """
    Electrical boresight (EB) steering directions for an electronically steered array.
    """

    _fields = ('DCXPoly', 'DCYPoly', 'UseEBPVP')
    _required = ('DCXPoly', 'DCYPoly')
    # descriptors
    DCXPoly = SerializableDescriptor(
        'DCXPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Electrical boresight steering *X-axis direction cosine (DCX)* as a function of '
                  'slow time ``(variable 1)``.')  # type: Poly1DType
    DCYPoly = SerializableDescriptor(
        'DCYPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Electrical boresight steering *Y-axis direction cosine (DCY)* as a function of '
                  'slow time ``(variable 1)``.')  # type: Poly1DType
    UseEBPVP = BooleanDescriptor(
        'UseEBPVP', _required, strict=DEFAULT_STRICT,
        docstring="")  # type: Optional[bool]

    def __init__(self, DCXPoly=None, DCYPoly=None, UseEBPVP=None, **kwargs):
        """
        Parameters
        ----------
        DCXPoly : Poly1DType|numpy.ndarray|list|tuple
        DCYPoly : Poly1DType|numpy.ndarray|list|tuple
        UseEBPVP : None|bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DCXPoly = DCXPoly
        self.DCYPoly = DCYPoly
        self.UseEBPVP = UseEBPVP
        super(EBType, self).__init__(**kwargs)

    def __call__(self, t):
        """
        Evaluate the polynomial at points `t`. This passes `t` straight through
        to :func:`polyval` of `numpy.polynomial.polynomial` for each of
        `DCXPoly,DCYPoly` components. If any of `DCXPoly,DCYPoly` is not populated,
        then `None` is returned.

        Parameters
        ----------
        t : float|int|numpy.ndarray
            The point(s) at which to evaluate.

        Returns
        -------
        None|numpy.ndarray
        """

        if self.DCXPoly is None or self.DCYPoly is None:
            return None
        return numpy.array([self.DCXPoly(t), self.DCYPoly(t)])

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.UseEBPVP is not None:
            required = max(required, (1, 1, 0))
        return required


class GainPhasePolyType(Serializable):
    """A container for the Gain and Phase Polygon definitions."""

    _fields = ('GainPoly', 'PhasePoly', 'AntGPId')
    _required = ('GainPoly', 'PhasePoly')
    # descriptors
    GainPoly = SerializableDescriptor(
        'GainPoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='One-way signal gain (in dB) as a function of X-axis direction cosine (DCX) (variable 1) '
                  'and Y-axis direction cosine (DCY) (variable 2). Gain relative to gain at DCX = 0 '
                  'and DCY = 0, so constant coefficient is always 0.0.')  # type: Poly2DType
    PhasePoly = SerializableDescriptor(
        'PhasePoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='One-way signal phase (in cycles) as a function of DCX (variable 1) and '
                  'DCY (variable 2). Phase relative to phase at DCX = 0 and DCY = 0, '
                  'so constant coefficient is always 0.0.')  # type: Poly2DType
    AntGPId = StringDescriptor(
        'AntGPId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of the Antenna Gain/Phase support array that specifies the '
                  'one-way pattern.')  # type: Optional[str]

    def __init__(self, GainPoly=None, PhasePoly=None, AntGPId=None, **kwargs):
        """
        Parameters
        ----------
        GainPoly : Poly2DType|numpy.ndarray|list|tuple
        PhasePoly : Poly2DType|numpy.ndarray|list|tuple
        AntGPId : None|str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GainPoly = GainPoly
        self.PhasePoly = PhasePoly
        self.AntGPId = AntGPId
        super(GainPhasePolyType, self).__init__(**kwargs)

    def __call__(self, x, y):
        """
        Evaluate a polynomial at points [`x`, `y`]. This passes `x`,`y` straight
        through to the call method for each component.

        Parameters
        ----------
        x : float|int|numpy.ndarray
            The first dependent variable of point(s) at which to evaluate.
        y : float|int|numpy.ndarray
            The second dependent variable of point(s) at which to evaluate.

        Returns
        -------
        numpy.ndarray
        """

        if self.GainPoly is None or self.PhasePoly is None:
            return None
        return numpy.array([self.GainPoly(x, y), self.PhasePoly(x, y)], dtype=numpy.float64)

    def minimize_order(self):
        """
        Trim the trailing zeros for each component coefficient array. This
        modifies the object in place.

        Returns
        -------
        None
        """

        self.GainPoly.minimize_order()
        self.PhasePoly.minimize_order()

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.AntGPId is not None:
            required = max(required, (1, 1, 0))
        return required


class AntPatternType(Serializable):
    """
    Parameter set that defines each Antenna Pattern as function time.
    """

    _fields = (
        'Identifier', 'FreqZero', 'GainZero', 'EBFreqShift', 'EBFreqShiftSF',
        'MLFreqDilation', 'MLFreqDilationSF', 'GainBSPoly', 'AntPolRef',
        'EB', 'Array', 'Element', 'GainPhaseArray')
    _required = (
        'Identifier', 'FreqZero', 'EB', 'Array', 'Element')
    _collections_tags = {
        'GainPhaseArray': {'array': False, 'child_tag': 'GainPhaseArray'}}
    _numeric_format = {'FreqZero': FLOAT_FORMAT, 'GainZero': FLOAT_FORMAT}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this ACF.')  # type: str
    FreqZero = FloatDescriptor(
        'FreqZero', _required, strict=DEFAULT_STRICT,
        docstring='The reference frequency value for which the Electrical Boresight '
                  'and array pattern polynomials are computed.')  # type: float
    GainZero = FloatDescriptor(
        'GainZero', _required, strict=DEFAULT_STRICT,
        docstring='The reference antenna gain at zero steering angle at the '
                  'reference frequency, measured in dB.')  # type: float
    EBFreqShift = BooleanDescriptor(
        'EBFreqShift', _required, strict=DEFAULT_STRICT,
        docstring="Parameter indicating whether the electronic boresite shifts with "
                  "frequency.")  # type: bool
    EBFreqShiftSF = SerializableDescriptor(
        'EBFreqShiftSF', FreqSFType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[FreqSFType]
    MLFreqDilation = BooleanDescriptor(
        'MLFreqDilation', _required, strict=DEFAULT_STRICT,
        docstring="Parameter indicating the mainlobe (ML) width changes with "
                  "frequency.")  # type: bool
    MLFreqDilationSF = SerializableDescriptor(
        'MLFreqDilationSF', FreqSFType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[FreqSFType]
    GainBSPoly = SerializableDescriptor(
        'GainBSPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Gain polynomial *(in dB)* as a function of frequency for boresight *(BS)* '
                  'at :math:`DCX=0, DCY=0`. '
                  'Frequency ratio :math:`(f-f0)/f0` is the input variable, and the constant '
                  'coefficient is always `0.0`.')  # type: Poly1DType
    AntPolRef = SerializableDescriptor(
        'AntPolRef', AntPolRefType, _required, strict=DEFAULT_STRICT,
        docstring='Polarization parameters for the EB steered to mechanical '
                  'boresight (EB_DCX = 0 and EB_DCY = 0).')  # type: AntPolRefType
    EB = SerializableDescriptor(
        'EB', EBType, _required, strict=DEFAULT_STRICT,
        docstring='Electrical boresight *(EB)* steering directions for an electronically '
                  'steered array.')  # type: EBType
    Array = SerializableDescriptor(
        'Array', GainPhasePolyType, _required, strict=DEFAULT_STRICT,
        docstring='Array pattern polynomials that define the shape of the '
                  'main-lobe.')  # type: GainPhasePolyType
    Element = SerializableDescriptor(
        'Element', GainPhasePolyType, _required, strict=DEFAULT_STRICT,
        docstring='Element array pattern polynomials for electronically steered '
                  'arrays.')  # type: GainPhasePolyType
    GainPhaseArray = SerializableListDescriptor(
        'GainPhaseArray', GainPhaseArrayType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Array of parameters that identify 2-D sampled Gain and Phase patterns at '
                  'single frequency value.')  # type: Union[None, List[GainPhaseArrayType]]

    def __init__(
            self,
            Identifier=None,
            FreqZero=None,
            GainZero=None,
            EBFreqShift=None,
            EBFreqShiftSF: Optional[FreqSFType] = None,
            MLFreqDilation=None,
            MLFreqDilationSF: Optional[FreqSFType] = None,
            GainBSPoly=None,
            AntPolRef: Union[None, AntPolRefType, numpy.ndarray, list, tuple] = None,
            EB=None,
            Array=None,
            Element=None,
            GainPhaseArray=None,
            **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        FreqZero : float
        GainZero : float
        EBFreqShift : bool
        EBFreqShiftSF : None|FreqSFType|numpy.ndarray|list|tuple
        MLFreqDilation : bool
        MLFreqDilationSF : None|FreqSFType|numpy.ndarray|list|tuple
        GainBSPoly : None|Poly1DType|numpy.ndarray|list|tuple
        AntPolRef : None|AntPolRefType|numpy.ndarray|list|tuple
        EB : EBType
        Array : GainPhasePolyType
        Element : GainPhasePolyType
        GainPhaseArray : None|List[GainPhaseArrayType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.FreqZero = FreqZero
        self.GainZero = GainZero
        self.EBFreqShift = EBFreqShift
        self.EBFreqShiftSF = EBFreqShiftSF
        self.MLFreqDilation = MLFreqDilation
        self.MLFreqDilationSF = MLFreqDilationSF
        self.GainBSPoly = GainBSPoly
        self.AntPolRef = AntPolRef
        self.EB = EB
        self.Array = Array
        self.Element = Element
        self.GainPhaseArray = GainPhaseArray
        super(AntPatternType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        for fld in ['EB', 'Array', 'Element']:
            val = getattr(self, fld)
            if val is not None:
                required = max(required, val.version_required())
        if self.EBFreqShiftSF is not None or \
                self.MLFreqDilationSF is not None or \
                self.AntPolRef is not None:
            required = max(required, (1, 1, 0))
        return required


class AntennaType(Serializable):
    """
    Parameters that describe the transmit and receive antennas used to collect
    the signal array(s).
    """

    _fields = (
        'NumACFs', 'NumAPCs', 'NumAntPats', 'AntCoordFrame', 'AntPhaseCenter', 'AntPattern')
    _required = _fields
    _collections_tags = {
        'AntCoordFrame': {'array': False, 'child_tag': 'AntCoordFrame'},
        'AntPhaseCenter': {'array': False, 'child_tag': 'AntPhaseCenter'},
        'AntPattern': {'array': False, 'child_tag': 'AntPattern'}}
    # descriptors
    AntCoordFrame = SerializableListDescriptor(
        'AntCoordFrame', AntCoordFrameType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Unit vectors that describe the orientation of an Antenna Coordinate Frame (ACF) '
                  'as function of time. Parameter set repeated for '
                  'each ACF.')  # type: List[AntCoordFrameType]
    AntPhaseCenter = SerializableListDescriptor(
        'AntPhaseCenter', AntPhaseCenterType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe each Antenna Phase Center (APC). Parameter '
                  'set repeated for each APC.')  # type: List[AntPhaseCenterType]
    AntPattern = SerializableListDescriptor(
        'AntPattern', AntPatternType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameter set that defines each Antenna Pattern as function time. Parameters '
                  'set repeated for each Antenna Pattern.')  # type: List[AntPatternType]

    def __init__(self, AntCoordFrame=None, AntPhaseCenter=None, AntPattern=None, **kwargs):
        """

        Parameters
        ----------
        AntCoordFrame : List[AntCoordFrameType]
        AntPhaseCenter : List[AntPhaseCenterType]
        AntPattern : List[AntPatternType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AntCoordFrame = AntCoordFrame
        self.AntPhaseCenter = AntPhaseCenter
        self.AntPattern = AntPattern
        super(AntennaType, self).__init__(**kwargs)

    @property
    def NumACFs(self):
        """
        int: The number of antenna coordinate frame elements.
        """

        if self.AntCoordFrame is None:
            return 0
        return len(self.AntCoordFrame)

    @property
    def NumAPCs(self):
        """
        int: The number of antenna phase center elements.
        """

        if self.AntPhaseCenter is None:
            return 0
        return len(self.AntPhaseCenter)

    @property
    def NumAntPats(self):
        """
        int: The number of antenna pattern elements.
        """

        if self.AntPattern is None:
            return 0
        return len(self.AntPattern)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.AntCoordFrame is not None:
            for entry in self.AntCoordFrame:
                required = max(required, entry.version_required())
        if self.AntPattern is not None:
            for entry in self.AntPattern:
                required = max(required, entry.version_required())

        return required



"""
Base NITF Header functionality definition.
"""
__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from weakref import WeakKeyDictionary
from typing import Union, List, Tuple
from collections import OrderedDict
import struct

import numpy

from sarpy.compliance import bytes_to_string
from .tres.registration import find_tre

logger = logging.getLogger(__name__)


# Base NITF type
class BaseNITFElement(object):

    @classmethod
    def minimum_length(cls):
        """
        The minimum size in bytes that takes to write this header element.

        Returns
        -------
        int
        """

        raise NotImplementedError

    def get_bytes_length(self):
        """
        Get the length of the serialized bytes array

        Returns
        -------
        int
        """

        raise NotImplementedError

    def to_bytes(self):
        """
        Write the object to a properly packed str.

        Returns
        -------
        bytes
        """

        raise NotImplementedError

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value: bytes|str
            the header string to scrape
        start : int
            the beginning location in the string

        Returns
        -------

        """

        raise NotImplementedError

    def to_json(self):
        """
        Serialize element to a json representation. This is intended to allow
        a simple presentation of the element.

        Returns
        -------
        dict
        """

        raise NotImplementedError


# Basic input and output interpreters

def _get_bytes(val, length):
    if val is None:
        return b''
    elif isinstance(val, int):
        frm_str = '{0:0' + str(length) + 'd}'
        return frm_str.format(val).encode('utf-8')
    elif isinstance(val, str):
        frm_str = '{0:' + str(length) + 's}'
        return frm_str.format(val).encode('utf-8')
    elif isinstance(val, bytes):
        if len(val) >= length:
            return val[:length]
        else:
            return val + b'\x00' * (length - len(val))
    else:
        raise TypeError('Unhandled type {}'.format(type(val)))


def _parse_int(val, length, default, name, instance):
    """
    Parse and/or validate the integer input.

    Parameters
    ----------
    val : None|int|bytes
    length : int
    default : None|int

    Returns
    -------
    int
    """

    if val is None:
        return default
    else:
        val = int(val)

    if -int(10)**(length-1) < val < int(10)**length:
        return val
    raise ValueError(
        'Integer {} cannot be rendered as a string of {} characters for '
        'attribute {} of class {}'.format(val, length, name, instance.__class__.__name__))


def _parse_float(val, default):
    """
    Sort of a special case.

    Parameters
    ----------
    val : None|float|bytes
    default : None|float

    Returns
    -------
    float
    """

    if val is None:
        return default
    else:
        return float(val)


def _parse_str(val, length, default, name, instance):
    """
    Parse and/or validate the string input.

    Parameters
    ----------
    val : None|str|bytes
    length : int
    default : None|str

    Returns
    -------
    str
    """

    if val is None:
        return default

    if isinstance(val, bytes):
        val = bytes_to_string(val)
    elif not isinstance(val, str):
        val = str(val)

    val = val.rstrip()
    if len(val) <= length:
        return val
    else:
        logger.warning(
            'Got string input value of length {} for attribute {} of class {}, '
            'which is longer than the allowed length {}, so '
            'truncating'.format(len(val), name, instance.__class__.__name__, length))
        return val[:length]


def _parse_bytes(val, length, default, name, instance):
    """
    Validate the raw/bytes input.

    Parameters
    ----------
    val : None|bytes
    length : int
    default : None|int

    Returns
    -------
    int
    """

    if val is None:
        return default
    elif isinstance(val, bytes):
        if len(val) <= length:
            return val
        else:
            logger.warning(
                'Got string input value of length {} for attribute {} of class {}, '
                'which is longer than the allowed length {}, so '
                'truncating'.format(len(val), name, instance.__class__.__name__, length))
            return val[:length]
    else:
        raise TypeError(
            'Expected type int or bytes for attribute {} of class {}, '
            'and got {}'.format(name, instance.__class__.__name__, type(val)))


def _parse_nitf_element(val, nitf_type, default_args, name, instance):
    if not issubclass(nitf_type, BaseNITFElement):
        raise TypeError(
            'nitf_type for attribute {} of class {} must be a subclass of '
            'BaseNITFElement'.format(name, nitf_type.__class__.__name__))

    if val is None:
        if default_args is None:
            return None
        return nitf_type(**default_args)
    elif isinstance(val, bytes):
        return nitf_type.from_bytes(val, 0)
    elif isinstance(val, nitf_type):
        return val
    else:
        raise ValueError(
            'Attribute {} for class {} requires an input of type bytes or {}. '
            'Got {}'.format(name, instance.__class__.__name__, nitf_type, type(val)))


# NITF Descriptors

class _BasicDescriptor(object):
    """A descriptor object for reusable properties. Note that is required that the calling instance is hashable."""
    _typ_string = None

    def __init__(self, name, required, length, docstring=''):
        self.data = WeakKeyDictionary()  # our instance reference dictionary
        # WeakDictionary use is subtle here. A reference to a particular class instance in this dictionary
        # should not be the thing keeping a particular class instance from being destroyed.
        self.name = name
        self.required = required
        self.length = length

        self.__doc__ = docstring
        self._format_docstring()

    def _format_docstring(self):
        docstring = self.__doc__
        if docstring is None:
            docstring = ''
        if (self._typ_string is not None) and (not docstring.startswith(self._typ_string)):
            docstring = '{} {}'.format(self._typ_string, docstring)

        suff = self._docstring_suffix()
        if suff is not None:
            docstring = '{} {}'.format(docstring, suff)

        if not self.required:
            docstring = '{} {}'.format(docstring, ' **Conditional.**')
        self.__doc__ = docstring

    def _docstring_suffix(self):
        return None

    def _get_default(self, instance):
        return None

    def __get__(self, instance, owner):
        """The getter.

        Parameters
        ----------
        instance : object
            the calling class instance
        owner : object
            the type of the class - that is, the actual object to which this descriptor is assigned

        Returns
        -------
        object
            the return value
        """

        if instance is None:
            # this has been access on the class, so return the class
            return self

        fetched = self.data.get(instance, None)
        if fetched is not None or not self.required:
            return fetched
        else:
            msg = 'Required field {} of class {} is not populated.'.format(self.name, instance.__class__.__name__)
            raise AttributeError(msg)

    def __set__(self, instance, value):
        """The setter method.

        Parameters
        ----------
        instance : object
            the calling class instance
        value
            the value to use in setting - the type depends of the specific extension of this base class

        Returns
        -------
        bool
            this base class, and only this base class, handles the required compliance and None behavior and has
            a return. This returns True if this the setting value was None, and False otherwise.
        """

        # NOTE: This is intended to handle this case for every extension of this class. Hence the boolean return,
        # which extensions SHOULD NOT implement. This is merely to follow DRY principles.
        if value is None:
            default_value = self._get_default(instance)
            if default_value is not None:
                self.data[instance] = default_value
                return True
            elif self.required:
                raise ValueError(
                    'Attribute {} of class {} cannot be assigned None.'.format(self.name, instance.__class__.__name__))
            self.data[instance] = None
            return True
        # note that the remainder must be implemented in each extension
        return False  # this is probably a bad habit, but this returns something for convenience alone


class _StringDescriptor(_BasicDescriptor):
    """A descriptor for string type"""
    _typ_string = 'str:'

    def __init__(self, name, required, length, default_value='', docstring=None):
        self._default_value = default_value
        super(_StringDescriptor, self).__init__(
            name, required, length, docstring=docstring)

    def _get_default(self, instance):
        return self._default_value

    def _docstring_suffix(self):
        if self._default_value is not None and len(self._default_value) > 0:
            return ' Default value is :code:`{}`.'.format(self._default_value)

    def __set__(self, instance, value):

        if super(_StringDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return
        self.data[instance] = _parse_str(value, self.length, self._default_value, self.name, instance)


class _StringEnumDescriptor(_BasicDescriptor):
    """A descriptor for enumerated (specified) string type.
    **The valid entries are case-sensitive and should be stripped of white space on each end.**"""
    _typ_string = 'str:'

    def __init__(self, name, required, length, values, default_value=None, docstring=None):
        self.values = values
        self._default_value = default_value
        super(_StringEnumDescriptor, self).__init__(
            name, required, length, docstring=docstring)
        if (self._default_value is not None) and (self._default_value not in self.values):
            self._default_value = None

    def _get_default(self, instance):
        return self._default_value

    def _docstring_suffix(self):
        suff = ' Takes values in :code:`{}`.'.format(self.values)
        if self._default_value is not None and len(self._default_value) > 0:
            suff += ' Default value is :code:`{}`.'.format(self._default_value)
        return suff

    def __set__(self, instance, value):
        if value is None:
            if self._default_value is not None:
                self.data[instance] = self._default_value
            else:
                super(_StringEnumDescriptor, self).__set__(instance, value)
            return

        val = _parse_str(value, self.length, self._default_value, self.name, instance)

        if val in self.values:
            self.data[instance] = val
        elif self._default_value is not None:
            msg = 'Attribute {} of class {} received {}, but values ARE REQUIRED to be ' \
                  'one of {}. It has been set to the default ' \
                  'value.'.format(self.name, instance.__class__.__name__, value, self.values)
            logger.error(msg)
            self.data[instance] = self._default_value
        else:
            msg = 'Attribute {} of class {} received {}, but values ARE REQUIRED to be ' \
                  'one of {}. This should be resolved, or it may cause unexpected ' \
                  'issues.'.format(self.name, instance.__class__.__name__, value, self.values)
            logger.error(msg)
            self.data[instance] = val


class _IntegerDescriptor(_BasicDescriptor):
    """A descriptor for integer type"""
    _typ_string = 'int:'

    def __init__(self, name, required, length, default_value=0, docstring=None):
        self._default_value = default_value
        super(_IntegerDescriptor, self).__init__(
            name, required, length, docstring=docstring)

    def _get_default(self, instance):
        return self._default_value

    def _docstring_suffix(self):
        if self._default_value is not None:
            return ' Default value is :code:`{}`.'.format(self._default_value)

    def __set__(self, instance, value):
        if super(_IntegerDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        iv = _parse_int(value, self.length, self._default_value, self.name, instance)
        self.data[instance] = iv


class _FloatDescriptor(_BasicDescriptor):
    """A special case descriptor for float type"""
    _typ_string = 'float:'

    def __init__(self, name, required, length, default_value=0, docstring=None):
        self._default_value = default_value
        super(_FloatDescriptor, self).__init__(
            name, required, length, docstring=docstring)

    def _get_default(self, instance):
        return self._default_value

    def _docstring_suffix(self):
        if self._default_value is not None:
            return ' Default value is :code:`{}`.'.format(self._default_value)

    def __set__(self, instance, value):
        if super(_FloatDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        iv = _parse_float(value, self._default_value)
        self.data[instance] = iv


class _RawDescriptor(_BasicDescriptor):
    """A descriptor for bytes type"""
    _typ_string = 'bytes:'

    def __init__(self, name, required, length, default_value=None, docstring=None):
        self._default_value = default_value
        super(_RawDescriptor, self).__init__(
            name, required, length, docstring=docstring)

    def _get_default(self, instance):
        return self._default_value

    def __set__(self, instance, value):
        if super(_RawDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        iv = _parse_bytes(value, self.length, self._default_value, self.name, instance)
        self.data[instance] = iv


class _NITFElementDescriptor(_BasicDescriptor):
    """A descriptor for properties of a specified type assumed to be an extension of Serializable"""

    def __init__(self, name, required, the_type, default_args=None, docstring=None):
        self.the_type = the_type
        self._typ_string = str(the_type).strip().split('.')[-1][:-2] + ':'
        self._default_args = default_args
        super(_NITFElementDescriptor, self).__init__(name, required, None, docstring=docstring)

    def _get_default(self, instance):
        if self._default_args is not None:
            return self.the_type(**self._default_args)
        return None

    def __set__(self, instance, value):
        if super(_NITFElementDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        self.data[instance] = _parse_nitf_element(value, self.the_type, self._default_args, self.name, instance)


# Concrete NITF element types

class NITFElement(BaseNITFElement):
    _ordering = ()
    _lengths = {}
    _binary_format = {}

    def __init__(self, **kwargs):
        for fld in self._ordering:
            # noinspection PyBroadException
            try:
                setattr(self, fld, kwargs.get(fld, None))
            except Exception:
                logger.critical('Failed setting attribute {} for class {}'.format(fld, self.__class__))
                raise

    @classmethod
    def minimum_length(cls):
        """
        The minimum size in bytes that takes to write this header element.

        Returns
        -------
        int
        """

        return sum(cls._lengths.values())

    def _get_attribute_length(self, fld):
        if fld not in self._ordering:
            return 0

        if fld in self._lengths:
            return self._lengths[fld]
        else:
            val = getattr(self, fld)
            if val is None:
                return 0
            elif isinstance(val, BaseNITFElement):
                return val.get_bytes_length()
            else:
                raise TypeError(
                    'Unhandled type {} for attribute {} of '
                    'class {}'.format(type(val), fld, self.__class__.__name__))

    def _get_attribute_bytes(self, fld):
        if fld not in self._ordering:
            return b''

        val = getattr(self, fld)
        if isinstance(val, BaseNITFElement):
            return val.to_bytes()
        elif fld in self._binary_format:
            return struct.pack(self._binary_format[fld], val)
        elif fld in self._lengths:
            return _get_bytes(val, self._lengths[fld])
        else:
            raise ValueError(
                'Unhandled attribute {} for class {}'.format(fld, self.__class__.__name__))

    def get_bytes_length(self):
        """
        Get the length of the serialized bytes array

        Returns
        -------
        int
        """

        return sum(self._get_attribute_length(fld) for fld in self._ordering)

    def to_bytes(self):
        """
        Write the object to a properly packed str.

        Returns
        -------
        bytes
        """

        return b''.join(self._get_attribute_bytes(fld) for fld in self._ordering)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        """

        Parameters
        ----------
        fields : dict
            The attribute:value dictionary.
        attribute : str
            The attribute name.
        value : bytes
            The bytes array to be parsed.
        start : int
            The present position in `value`.

        Returns
        -------
        int
            The position in `value` after parsing this attribute.
        """

        if attribute not in cls._ordering:
            raise ValueError('Unexpected attribute {}'.format(attribute))
        if attribute in fields:
            return start
        if attribute in cls._binary_format:
            if attribute not in cls._lengths:
                raise ValueError(
                    'attribute {} has binary format specified, but no length specified '
                    'for class {}'.format(attribute, cls))
            end = start + cls._lengths[attribute]
            fields[attribute] = struct.unpack(cls._binary_format[attribute], value[start:end])[0]
            return end
        elif attribute in cls._lengths:
            end = start + cls._lengths[attribute]
            fields[attribute] = value[start:end]
            return end
        elif hasattr(cls, attribute):
            the_typ = getattr(cls, attribute).the_type
            assert issubclass(the_typ, BaseNITFElement)
            the_value = the_typ.from_bytes(value, start)
            fields[attribute] = the_value
            return start + the_value.get_bytes_length()
        else:
            raise ValueError('Cannot parse attribute {} for class {}'.format(attribute, cls))

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value: bytes|str
            the header string to scrape
        start : int
            the beginning location in the string

        Returns
        -------

        """

        fields = {}
        loc = start
        for fld in cls._ordering:
            loc = cls._parse_attribute(fields, fld, value, loc)
        return cls(**fields)

    def to_json(self):
        out = OrderedDict()
        for fld in self._ordering:
            if self._get_attribute_length(fld) == 0:
                continue
            value = getattr(self, fld)
            if value is None:
                out[fld] = ''
            elif isinstance(value, (str, bytes, int)):
                out[fld] = value
            elif isinstance(value, BaseNITFElement):
                out[fld] = value.to_json()
            else:
                logger.error(
                    'Got unhandled type `{}` for json serialization for '
                    'attribute `{}` of class {}'.format(type(value), fld, self.__class__))
        return out


class NITFLoop(NITFElement):
    __slots__ = ('_values', )
    _ordering = ('values', )
    _child_class = None  # must be a subclass of NITFElement
    _count_size = 0  # type: int

    def __init__(self, values=None, **kwargs):
        if not issubclass(self._child_class, NITFElement):
            raise TypeError('_child_class for {} must be a subclass of NITFElement'.format(self.__class__.__name__))
        self._values = tuple()
        super(NITFLoop, self).__init__(values=values, **kwargs)

    @property
    def values(self):
        # type: () -> Tuple[_child_class, ...]
        return self._values

    @values.setter
    def values(self, value):
        if value is None:
            self._values = ()
            return
        if not isinstance(value, tuple):
            value = tuple(value)
        for i, entry in enumerate(value):
            if not isinstance(entry, self._child_class):
                raise TypeError(
                    'values must be of type {}, got entry {} of type {}'.format(
                        self._child_class, i, type(entry)))
        self._values = value

    def __len__(self):
        return len(self._values)

    def __getitem__(self, item):
        # type: (Union[int, slice]) -> Union[_child_class, List[_child_class]]
        return self._values[item]

    def get_bytes_length(self):
        return self._count_size + sum(entry.get_bytes_length() for entry in self._values)

    @classmethod
    def minimum_length(cls):
        return cls._count_size

    @classmethod
    def _parse_count(cls, value, start):
        loc = start
        count = int(value[loc:loc+cls._count_size])
        loc += cls._count_size
        return count, loc

    @classmethod
    def from_bytes(cls, value, start):
        if not issubclass(cls._child_class, NITFElement):
            raise TypeError('_child_class for {} must be a subclass of NITFElement'.format(cls.__name__))

        count, loc = cls._parse_count(value, start)
        if count == 0:
            return cls(values=None)

        values = []
        for i in range(count):
            val = cls._child_class.from_bytes(value, loc)
            loc += val.get_bytes_length()
            values.append(val)
        return cls(values=values)

    def _counts_bytes(self):
        frm_str = '{0:0'+str(self._count_size) + 'd}'
        return frm_str.format(len(self.values)).encode('utf-8')

    def to_bytes(self):
        return self._counts_bytes() + b''.join(entry.to_bytes() for entry in self._values)

    def to_json(self):
        return [entry.to_json() for entry in self._values]


class Unstructured(NITFElement):
    """
    A possible NITF element pattern which is largely unparsed -
    just a bytes array of a given length
    """

    __slots__ = ('_data', )
    _ordering = ('data', )
    _size_len = 1

    def __init__(self, data=None, **kwargs):
        self._data = None
        if not (isinstance(self._size_len, int) and self._size_len > 0):
            raise TypeError(
                'class variable _size_len for {} must be a positive '
                'integer'.format(self.__class__.__name__))
        super(Unstructured, self).__init__(data=data, **kwargs)

    @property
    def data(self):  # type: () -> Union[None, bytes, NITFElement, TREList]
        return self._data

    @data.setter
    def data(self, value):
        if value is None:
            self._data = None
            return

        if not isinstance(value, (bytes, NITFElement)):
            raise TypeError(
                'data requires bytes or NITFElement type. '
                'Got type {}'.format(type(value)))
        siz_lim = 10**self._size_len - 1
        if isinstance(value, bytes):
            len_cond = (len(value) > siz_lim)
        else:
            len_cond = value.get_bytes_length() > siz_lim
        if len_cond:
            raise ValueError('The provided data is longer than {}'.format(siz_lim))
        self._data = value
        self._populate_data()

    def _populate_data(self):
        """
        Populate the _data attribute from bytes to some other appropriate object.

        Returns
        -------
        None
        """

        pass

    @classmethod
    def minimum_length(cls):
        return cls._size_len

    def _get_attribute_bytes(self, attribute):
        if attribute == 'data':
            siz_frm = '{0:0' + str(self._size_len) + '}'
            data = self.data
            if data is None:
                return b'0'*self._size_len
            if isinstance(data, NITFElement):
                data = data.to_bytes()
            if isinstance(data, bytes):
                return siz_frm.format(len(data)).encode('utf-8') + data
            else:
                raise TypeError(
                    'Got unexpected data type {} for attribute {} of class {}'.format(
                        type(data), attribute, self.__class__))
        return super(Unstructured, self)._get_attribute_bytes(attribute)

    def _get_attribute_length(self, attribute):
        if attribute == 'data':
            data = self.data
            if data is None:
                return self._size_len
            elif isinstance(data, NITFElement):
                return self._size_len + data.get_bytes_length()
            else:
                return self._size_len + len(data)
        return super(Unstructured, self)._get_attribute_length(attribute)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'data':
            length = int(value[start:start + cls._size_len])
            start += cls._size_len
            fields['data'] = value[start:start + length]
            return start + length
        return super(Unstructured, cls)._parse_attribute(fields, attribute, value, start)


class _ItemArrayHeaders(BaseNITFElement):
    """
    Item array in the NITF header (i.e. Image Segment, Text Segment).
    This is not really meant to be used directly.
    """

    __slots__ = ('subhead_sizes', 'item_sizes')
    _subhead_len = 0
    _item_len = 0

    def __init__(self, subhead_sizes=None, item_sizes=None, **kwargs):
        """

        Parameters
        ----------
        subhead_sizes : numpy.ndarray|None
        item_sizes : numpy.ndarray|None
        """

        if subhead_sizes is None or item_sizes is None:
            subhead_sizes = numpy.zeros((0, ), dtype=numpy.int64)
            item_sizes = numpy.zeros((0,), dtype=numpy.int64)
        if subhead_sizes.shape != item_sizes.shape or len(item_sizes.shape) != 1:
            raise ValueError(
                'the subhead_offsets and item_offsets arrays must one-dimensional and the same length')

        self.subhead_sizes = subhead_sizes
        """
        numpy.ndarray: the subheader sizes
        """

        self.item_sizes = item_sizes
        """
        numpy.ndarray: the item size
        """

        super(_ItemArrayHeaders, self).__init__(**kwargs)

    def get_bytes_length(self):
        return 3 + (self._subhead_len + self._item_len)*self.subhead_sizes.size

    @classmethod
    def minimum_length(cls):
        return 3

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes|str
        start : int

        Returns
        -------
        _ItemArrayHeaders
        """

        subhead_len, item_len = int(cls._subhead_len), int(cls._item_len)
        if len(value) < start + 3:
            raise ValueError('value must have length at least {}. Got {}'.format(start+3, len(value)))
        loc = start
        count = int(value[loc:loc+3])
        length = 3 + count*(subhead_len + item_len)
        if len(value) < start + length:
            raise ValueError('value must have length at least {}. Got {}'.format(start+length, len(value)))
        loc += 3
        subhead_sizes = numpy.zeros((count, ), dtype=numpy.int64)
        item_sizes = numpy.zeros((count, ), dtype=numpy.int64)
        for i in range(count):
            subhead_sizes[i] = int(value[loc: loc+subhead_len])
            loc += subhead_len
            item_sizes[i] = int(value[loc: loc+item_len])
            loc += item_len
        return cls(subhead_sizes, item_sizes)

    def to_bytes(self):
        out = '{0:03d}'.format(self.subhead_sizes.size)
        subh_frm = '{0:0' + str(self._subhead_len) + 'd}'
        item_frm = '{0:0' + str(self._item_len) + 'd}'
        for sh_off, it_off in zip(self.subhead_sizes, self.item_sizes):
            out += subh_frm.format(sh_off) + item_frm.format(it_off)
        return out.encode('utf-8')

    def to_json(self):
        return OrderedDict([
            ('subheader_sizes', self.subhead_sizes.tolist()),
            ('item_sizes', self.item_sizes.tolist())])


######
# TRE Elements

class TRE(BaseNITFElement):
    """
    An abstract TRE class - this should not be instantiated directly.
    """

    @property
    def TAG(self):
        """
        str: The TRE tag.
        """

        raise NotImplementedError

    @property
    def DATA(self):
        """
        The TRE data.
        """

        raise NotImplementedError

    @property
    def EL(self):
        """
        int: The TRE element length.
        """

        raise NotImplementedError

    def get_bytes_length(self):
        raise NotImplementedError

    def to_bytes(self):
        raise NotImplementedError

    @classmethod
    def minimum_length(cls):
        return 11

    @classmethod
    def from_bytes(cls, value, start):
        tag = value[start:start+6]
        known_tre = find_tre(tag)
        if known_tre is not None:
            try:
                return known_tre.from_bytes(value, start)
            except Exception as e:
                logger.error(
                    "Returning unparsed tre, because we failed parsing tre as "
                    "type {} with exception\n\t{}".format(known_tre.__name__, e))
        return UnknownTRE.from_bytes(value, start)

    def to_json(self):
        out = OrderedDict([('tag', self.TAG), ('length', self.EL)])
        if isinstance(self.DATA, bytes):
            out['data'] = self.DATA
        else:
            out['data'] = self.DATA.to_json()
        return out


class UnknownTRE(TRE):
    __slots__ = ('_TAG', '_data')

    def __init__(self, TAG, data):
        """

        Parameters
        ----------
        TAG : str
        data : bytes
        """

        self._data = None
        if isinstance(TAG, bytes):
            TAG = TAG.decode('utf-8')

        if not isinstance(TAG, str):
            raise TypeError('TAG must be a string. Got {}'.format(type(TAG)))
        if len(TAG) > 6:
            raise ValueError('TAG must be 6 or fewer characters')

        self._TAG = TAG
        self._data = data

    @property
    def TAG(self):
        return self._TAG

    @property
    def DATA(self):  # type: () -> bytes
        return self._data

    @DATA.setter
    def DATA(self, value):
        if not isinstance(value, bytes):
            raise TypeError('data must be a bytes instance. Got {}'.format(type(value)))
        self._data = value

    @property
    def EL(self):
        return len(self._data)

    def get_bytes_length(self):
        return 11 + self.EL

    def to_bytes(self):
        return '{0:6s}{1:05d}'.format(self.TAG, self.EL).encode('utf-8') + self._data

    @classmethod
    def from_bytes(cls, value, start):
        tag = value[start:start+6]
        lng = int(value[start+6:start+11])
        return cls(tag, value[start+11:start+11+lng])


class TREList(NITFElement):
    """
    A list of TREs. This is meant to be used indirectly through one of the header
    type objects, which controls the parsing appropriately.
    """

    __slots__ = ('_tres', )
    _ordering = ('tres', )

    def __init__(self, tres=None, **kwargs):
        self._tres = []
        super(TREList, self).__init__(tres=tres, **kwargs)

    @property
    def tres(self):
        # type: () -> List[TRE]
        return self._tres

    @tres.setter
    def tres(self, value):
        if value is None:
            self._tres = []
            return

        if not isinstance(value, (list, tuple)):
            raise TypeError('tres must be a list or tuple')

        for i, entry in enumerate(value):
            if not isinstance(entry, TRE):
                raise TypeError(
                    'Each entry of tres must be of type TRE. '
                    'Entry {} is type {}'.format(i, type(entry)))
        self._tres = value

    def _get_attribute_bytes(self, attribute):
        if attribute == 'tres':
            if len(self._tres) == 0:
                return b''
            return b''.join(entry.to_bytes() for entry in self._tres)
        return super(TREList, self)._get_attribute_bytes(attribute)

    def _get_attribute_length(self, attribute):
        if attribute == 'tres':
            if len(self._tres) == 0:
                return 0
            return sum(entry.get_bytes_length() for entry in self._tres)
        return super(TREList, self)._get_attribute_length(attribute)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'tres':
            if len(value) == start:
                fields['tres'] = []
                return start
            tres = []
            loc = start
            while loc < len(value):
                anticipated_length = int(value[loc+6:loc+11]) + 11
                tre = TRE.from_bytes(value, loc)
                parsed_length = tre.get_bytes_length()
                if parsed_length != anticipated_length:
                    logger.error(
                        'The given length for TRE {} instance is {}, but the constructed length is {}. '
                        'This is the result of a malformed TRE object definition. '
                        'If possible, this should be reported to the sarpy team for review/repair.'.format(
                            tre.TAG, anticipated_length, parsed_length))
                loc += anticipated_length
                tres.append(tre)
            fields['tres'] = tres
            return len(value)
        return super(TREList, cls)._parse_attribute(fields, attribute, value, start)

    def __len__(self):
        return len(self._tres)

    def __getitem__(self, item):
        # type: (Union[int, slice, str]) -> Union[None, TRE, List[TRE]]
        if isinstance(item, (int, slice)):
            return self._tres[item]
        elif isinstance(item, str):
            for entry in self.tres:
                if entry.TAG == item:
                    return entry
            return None
        else:
            raise TypeError('Got unhandled type {}'.format(type(item)))

    def to_json(self):
        return [entry.to_json() for entry in self._tres]


class TREHeader(Unstructured):
    def _populate_data(self):
        if isinstance(self._data, bytes):
            data = TREList.from_bytes(self._data, 0)
            self._data = data


class UserHeaderType(Unstructured):
    __slots__ = ('_data', '_ofl')
    _ordering = ('data', )
    _size_len = 5
    _ofl_len = 3

    def __init__(self, OFL=None, data=None, **kwargs):
        self._ofl = None
        self._data = None
        self.OFL = OFL
        super(UserHeaderType, self).__init__(data=data, **kwargs)

    @property
    def OFL(self):  # type: () -> int
        return self._ofl

    @OFL.setter
    def OFL(self, value):
        if value is None:
            self._ofl = 0
            return

        value = int(value)
        if not (0 <= value <= 999):
            raise ValueError('ofl requires an integer value in the range 0-999.')
        self._ofl = value

    def _populate_data(self):
        if isinstance(self._data, bytes):
            data = TREList.from_bytes(self._data, 0)
            self._data = data

    @classmethod
    def minimum_length(cls):
        return cls._size_len

    def _get_attribute_bytes(self, attribute):
        if attribute == 'data':
            siz_frm = '{0:0' + str(self._size_len) + 'd}'
            ofl_frm = '{0:0' + str(self._ofl_len) + 'd}'
            data = self.data
            if data is None:
                return b'0'*self._size_len
            if isinstance(data, NITFElement):
                data = data.to_bytes()
            if isinstance(data, bytes):
                return siz_frm.format(len(data) + self._ofl_len).encode('utf-8') + \
                    ofl_frm.format(self._ofl).encode('utf-8') + data
            else:
                raise TypeError('Got unexpected data type {}'.format(type(data)))
        return super(Unstructured, self)._get_attribute_bytes(attribute)

    def _get_attribute_length(self, attribute):
        if attribute == 'data':
            data = self.data
            if data is None:
                return self._size_len
            elif isinstance(data, NITFElement):
                return self._size_len + self._ofl_len + data.get_bytes_length()
            else:
                return self._size_len + self._ofl_len + len(data)
        return super(UserHeaderType, self)._get_attribute_length(attribute)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'data':
            length = int(value[start:start + cls._size_len])
            start += cls._size_len
            if length > 0:
                ofl = int(value[start:start+cls._ofl_len])
                fields['OFL'] = ofl
                fields['data'] = value[start+cls._ofl_len:start + length]
            else:
                fields['OFL'] = 0
                fields['data'] = None
            return start + length
        return super(UserHeaderType, cls)._parse_attribute(fields, attribute, value, start)



"""
This package contains the CPHD schema
"""

__classification__ = 'UNCLASSIFIED'
__author__ = "Thomas McCullough"

import os
import re
from typing import List, Dict, Tuple, Union


_CPHD_DEFAULT_TUPLE = (1, 1, 0)

_the_directory = os.path.split(__file__)[0]

urn_mapping = {
    'urn:CPHD:0.3.0': {
        'tuple': (0, 3, 0),
        'version': '0.3.0',
        'release': '0.3.0',
        'date': ''},
    'urn:CPHD:1.0.1': {
        'tuple': (1, 0, 1),
        'version': '1.0.1',
        'release': '1.0.1',
        'date': '2018-05-21T00:00:00Z',
        'schema': os.path.join(_the_directory, 'CPHD_schema_V1.0.1_2018_05_21.xsd')},
    'urn:CPHD:1.1.0': {
        'tuple': (1, 1, 0),
        'version': '1.1.0',
        'release': '1.1.0',
        'date': '2021-11-30T00:00:00Z',
        'schema': os.path.join(_the_directory, 'CPHD_schema_V1.1.0_2021_11_30_FINAL.xsd')},
}
WRITABLE_VERSIONS = ('1.0.1', '1.1.0')

# validate the defined paths
for key, entry in urn_mapping.items():
    schema_path = entry.get('schema', None)
    if schema_path is not None and not os.path.exists(schema_path):
        raise ValueError('`{}` has nonexistent schema path {}'.format(key, schema_path))


def get_default_tuple() -> Tuple[int, int, int]:
    """
    Get the default CPHD version tuple.

    Returns
    -------
    Tuple[int, int, int]
    """

    return _CPHD_DEFAULT_TUPLE


def get_default_version_string() -> str:
    """
    Get the default CPHD version string.

    Returns
    -------
    str
    """

    return '{}.{}.{}'.format(*_CPHD_DEFAULT_TUPLE)


def get_namespace(version: Union[str, Tuple[int, int, int]]) -> str:
    if isinstance(version, (list, tuple)):
        version = '{}.{}.{}'.format(version[0], version[1], version[2])
    return 'http://api.nsgreg.nga.mil/schema/cphd/{}'.format(version)


def check_urn(urn_string: str) -> str:
    """
    Checks that the urn string follows the correct pattern.

    Parameters
    ----------
    urn_string : str

    Raises
    ------
    ValueError
        This raises an exception for a poorly formed or unmapped CPHD urn.
    """

    if not isinstance(urn_string, str):
        raise TypeError(
            'Expected a urn input of string type, got type {}'.format(type(urn_string)))

    the_match = re.match(r'^\d.\d.\d$', urn_string)
    if the_match is not None:
        urn_string = 'urn:CPHD:{}'.format(urn_string)

    the_match = re.match(r'^urn:CPHD:\d.\d.\d$', urn_string)
    if the_match is None:
        raise ValueError(
            'Input provided as `{}`,\nbut should be of the form '
            '`urn:CPHD:<major>.<minor>.<release>'.format(urn_string))
    return urn_string


def get_urn_details(urn_string: str) -> Dict[str, str]:
    """
    Gets the associated details for the given CPHD urn, or raise an exception for
    poorly formatted or unrecognized urn.

    Parameters
    ----------
    urn_string : str

    Returns
    -------
    Dict[str, str]
    """

    urn_string = check_urn(urn_string)
    out = urn_mapping.get(urn_string, None)

    if out is None:
        raise KeyError(
            'Got correctly formatted, but unmapped CPHD urn {}.'.format(urn_string))
    return out


def get_schema_path(the_urn: str) -> str:
    """
    Gets the path to the proper schema file for the given urn.

    Parameters
    ----------
    the_urn : str

    Returns
    -------
    str
    """

    result = get_urn_details(the_urn)
    return result.get('schema', None)


def get_versions() -> List[str]:
    """
    Gets a list of recognized CPHD urns.

    Returns
    -------
    List[str]
    """

    return list(sorted(urn_mapping.keys(), key=lambda x: urn_mapping[x]['tuple']))



"""
The symbol header element definition - only applies to NITF 2.0
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import struct
import numpy

from .base import NITFElement, UserHeaderType, _IntegerDescriptor,\
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor
from .security import NITFSecurityTags0


class SymbolSegmentHeader(NITFElement):
    """
    Symbol segment subheader for NITF version 2.0 - see standards document
    MIL-STD-2500A for more information.
    """
    _ordering = (
        'SY', 'SID', 'SNAME', 'Security', 'ENCRYP', 'STYPE',
        'NLIPS', 'NPIXPL', 'NWDTH', 'NBPP', 'SDLVL', 'SALVL', 'SLOC',
        'SLOC2', 'SCOLOR', 'SNUM', 'SROT', 'DLUT', 'UserHeader')
    _lengths = {
        'SY': 2, 'SID': 10, 'SNAME': 20, 'ENCRYP': 1, 'STYPE': 1,
        'NLIPS': 4, 'NPIXPL': 4, 'NWDTH': 4, 'NBPP': 1,
        'SDLVL': 3, 'SALVL': 3, 'SLOC': 10, 'SLOC2': 10, 'SCOLOR': 1,
        'SNUM': 6, 'SROT': 3}
    SY = _StringEnumDescriptor('SY', True, 2, {'SY', })  # type: str
    SID = _StringDescriptor('SID', True, 10)  # type: str
    SNAME = _StringDescriptor('SNAME', True, 20)  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags0, default_args={})  # type: NITFSecurityTags0
    ENCRYP = _StringEnumDescriptor('ENCRYP', True, 1, {'0'})  # type: str
    STYPE = _StringDescriptor('STYPE', True, 1)  # type: str
    NLIPS = _IntegerDescriptor('NLIPS', True, 4)  # type: int
    NPIXPL = _IntegerDescriptor('NPIXPL', True, 4)  # type: int
    NWDTH = _IntegerDescriptor('NWDTH', True, 4)  # type: int
    NBPP = _IntegerDescriptor('NBPP', True, 1)  # type: int
    SDLVL = _IntegerDescriptor('SDLVL', True, 3)  # type: int
    SALVL = _IntegerDescriptor('SALVL', True, 3)  # type: int
    SLOC = _StringDescriptor('SLOC', True, 10)  # type: str
    SLOC2 = _StringDescriptor('SLOC2', True, 10)  # type: str
    SCOLOR = _StringDescriptor('SCOLOR', True, 1)  # type: str
    SNUM = _StringDescriptor('SLOC2', True, 6)  # type: str
    SROT = _IntegerDescriptor('SROT', True, 3)  # type: int
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={})  # type: UserHeaderType

    def __init__(self, **kwargs):
        self._DLUT = None
        super(SymbolSegmentHeader, self).__init__(**kwargs)

    @classmethod
    def minimum_length(cls):
        return 13

    @property
    def DLUT(self):
        """
        The Look-up Table (LUT) data.

        Returns
        -------
        None|numpy.ndarray
        """

        return self._DLUT

    @DLUT.setter
    def DLUT(self, value):
        if value is None:
            self._DLUT = None
            return

        if not isinstance(value, numpy.ndarray):
            raise TypeError('DLUT must be a numpy array')
        if value.dtype.name != 'uint8':
            raise ValueError('DLUT must be a numpy array of dtype uint8, got {}'.format(value.dtype.name))
        if value.ndim != 2 or value.shape[1] != 3:
            raise ValueError('DLUT must be a two-dimensional array of shape (N, 3).')
        if value.size > 256:
            raise ValueError(
                'The number of DLUT elements must be 256 or fewer. '
                'Got DLUT shape {}'.format(value.shape))
        self._DLUT = value

    @property
    def NELUT(self):
        """
        Number of LUT Entries.

        Returns
        -------
        int
        """

        return 0 if self._DLUT is None else self._DLUT.size

    def _get_attribute_bytes(self, attribute):
        if attribute == 'DLUT':
            if self.NELUT == 0:
                out = b'000'
            else:
                out = '{0:d}'.format(self.NELUT).encode() + \
                      struct.pack('{}B'.format(self.NELUT*3), *self.DLUT.flatten())
            return out
        else:
            return super(SymbolSegmentHeader, self)._get_attribute_bytes(attribute)

    def _get_attribute_length(self, attribute):
        if attribute == 'DLUT':
            return 3 + self.NELUT*3
        else:
            return super(SymbolSegmentHeader, self)._get_attribute_length(attribute)

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'DLUT':
            loc = start
            nelut = int(value[loc:loc + 3])
            loc += 3
            if nelut == 0:
                fields['DLUT'] = None
            else:
                fields['DLUT'] = numpy.array(
                    struct.unpack(
                        '{}B'.format(3*nelut),
                        value[loc:loc + 3*nelut]), dtype=numpy.uint8).reshape((nelut, 3))
                loc += nelut*3
            return loc
        return super(SymbolSegmentHeader, cls)._parse_attribute(fields, attribute, value, start)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class SENSRAType(TREElement):
    def __init__(self, value):
        super(SENSRAType, self).__init__()
        self.add_field('REFROW', 's', 8, value)
        self.add_field('REFCOL', 's', 8, value)
        self.add_field('SNSMODEL', 's', 6, value)
        self.add_field('SNSMOUNT', 's', 3, value)
        self.add_field('SENSLOC', 's', 21, value)
        self.add_field('SNALTSRC', 's', 1, value)
        self.add_field('SENSALT', 's', 6, value)
        self.add_field('SNALUNIT', 's', 1, value)
        self.add_field('SENSAGL', 's', 5, value)
        self.add_field('SNSPITCH', 's', 7, value)
        self.add_field('SENSROLL', 's', 8, value)
        self.add_field('SENSYAW', 's', 8, value)
        self.add_field('PLTPITCH', 's', 7, value)
        self.add_field('PLATROLL', 's', 8, value)
        self.add_field('PLATHDG', 's', 5, value)
        self.add_field('GRSPDSRC', 's', 1, value)
        self.add_field('GRDSPEED', 's', 6, value)
        self.add_field('GRSPUNIT', 's', 1, value)
        self.add_field('GRDTRACK', 's', 5, value)
        self.add_field('VERTVEL', 's', 5, value)
        self.add_field('VERTVELU', 's', 1, value)
        self.add_field('SWATHFRM', 's', 4, value)
        self.add_field('NSWATHS', 's', 4, value)
        self.add_field('SPOTNUM', 's', 3, value)


class SENSRA(TREExtension):
    _tag_value = 'SENSRA'
    _data_type = SENSRAType
